import re
from collections import defaultdict

_NEST_SCOPES = {'[': ']', '(': ')', '{': '}'}
_STR_SCOPES = '"' + "'"

#: Regex to find function calls.
_fre = re.compile(r"(\w+)\s?\((.*)\)")


def rewrite(text, kwfmt='_kw_{}'):
    """Rewrite all occurrences of functions calls adding a keyword to each
    non-keyword arg after a keyword arg.

    The generating format for the keyword can be specified using the kwfmt argument
    where the place holder is formatted to a number starting from 0.
    """
    _repl = lambda mo: repl(mo, kwfmt)
    return _fre.sub(_repl, text)


def repl(match):
    """Helper function for rewrite."""

    func, args = match.groups()

    fmt = kwfmt + '={}'
    scopes = defaultdict(int)
    ret = []

    n = -1
    found_first_kw = False
    current, has_kw = '', False

    for c in args + ',':
        in_scope = any(scopes.values())
        in_string = any(scopes[k] for k in _STR_SCOPES)

        if c in _NEST_SCOPES.keys():
            scopes[_NEST_SCOPES[c]] += 1

        elif c in _NEST_SCOPES.values() and not in_string:
            scopes[c] -= 1

        elif c in _STR_SCOPES:
            scopes[c] = not scopes[c]

        elif c == '=' and not in_scope:
            has_kw = True
            found_first_kw = True
            n = -1

        elif c == ',' and not in_scope:
            if current:
                current = rewrite(current.strip()).strip()
                if has_kw or not found_first_kw:
                    ret.append(current)
                else:
                    ret.append(fmt.format(n, current))

            n += 1
            current, has_kw = '', False
            continue

        current += c

    return '{}({})'.format(func, ', '.join(ret))

assert rewrite('f()') == 'f()'
assert rewrite('f1()') == 'f1()'
assert rewrite('fun32()') == 'fun32()'
assert rewrite('f ()') == 'f()'
assert rewrite('f(x)') == 'f(x)'
assert rewrite('f(len(s))') == 'f(len(s))'
assert rewrite('f(k=1)') == 'f(k=1)'
assert rewrite('f(k=")")') == 'f(k=")")'
assert rewrite('f(k=(1, 2, 3))') == 'f(k=(1, 2, 3))'
assert rewrite('f(k=1, y)') == 'f(k=1, _kw_0=y)'
assert rewrite('f(1, k=1, y)') == 'f(1, k=1, _kw_0=y)'
assert rewrite('f(g(k=1, y))') == 'f(g(k=1, _kw_0=y))'
assert rewrite('w = f(k=1, y)') == 'w = f(k=1, _kw_0=y)'
assert rewrite('with f(k=1, y) as fp') == 'with f(k=1, _kw_0=y) as fp'
assert rewrite('with f(k=1, y) as fp\n    w = f(k=1, y)') == 'with f(k=1, _kw_0=y) as fp\n    w = f(k=1, _kw_0=y)'



import os
from ayrton.parser.pyparser import parser, pytoken, metaparser

class PythonGrammar(parser.Grammar):

    KEYWORD_TOKEN = pytoken.python_tokens["NAME"]
    TOKENS = pytoken.python_tokens
    OPERATOR_MAP = pytoken.python_opmap

def _get_python_grammar():
    here = os.path.dirname(__file__)
    fp = open(os.path.join(here, "data", "Grammar3.3"))
    try:
        gram_source = fp.read()
    finally:
        fp.close()
    pgen = metaparser.ParserGenerator(gram_source)
    return pgen.build_grammar(PythonGrammar)


python_grammar = _get_python_grammar()

class _Tokens(object):
    pass
for tok_name, idx in pytoken.python_tokens.items():
    setattr(_Tokens, tok_name, idx)
tokens = _Tokens()

class _Symbols(object):
    pass
for sym_name, idx in python_grammar.symbol_ids.items():
    setattr(_Symbols, sym_name, idx)
syms = _Symbols()

del _get_python_grammar, _Tokens, tok_name, sym_name, idx



from ayrton.parser.pyparser import automata
from ayrton.parser.pyparser.pygram import tokens
from ayrton.parser.pyparser.pytoken import python_opmap
from ayrton.parser.pyparser.error import TokenError, TokenIndentationError, TabError
from ayrton.parser.pyparser.pytokenize import tabsize, alttabsize, whiteSpaceDFA, \
    triple_quoted, endDFAs, single_quoted, pseudoDFA
from ayrton.parser.astcompiler import consts

NAMECHARS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_'
NUMCHARS = '0123456789'
ALNUMCHARS = NAMECHARS + NUMCHARS
EXTENDED_ALNUMCHARS = ALNUMCHARS + '-.'
WHITESPACES = ' \t\n\r\v\f'

def match_encoding_declaration(comment):
    """returns the declared encoding or None

    This function is a replacement for :
    >>> py_encoding = re.compile(r"coding[:=]\s*([-\w.]+)")
    >>> py_encoding.search(comment)
    """
    index = comment.find('coding')
    if index < 0:
        return None
    next_char = comment[index + 6]
    if next_char not in ':=':
        return None
    end_of_decl = comment[index + 7:]
    index = 0
    for char in end_of_decl:
        if char not in WHITESPACES:
            break
        index += 1
    else:
        return None
    encoding = ''
    for char in end_of_decl[index:]:
        if char in EXTENDED_ALNUMCHARS:
            encoding += char
        else:
            break
    if encoding != '':
        return encoding
    return None


def verify_identifier(token):
    for c in token:
        if ord(c) > 0x80:
            break
    else:
        return True
    try:
        u = token.decode('utf-8')
    except UnicodeDecodeError:
        return False
    from pypy.objspace.std.unicodeobject import _isidentifier
    return _isidentifier(u)


def generate_tokens(lines, flags):
    """
    This is a rewrite of pypy.module.parser.pytokenize.generate_tokens since
    the original function is not RPYTHON (uses yield)
    It was also slightly modified to generate Token instances instead
    of the original 5-tuples -- it's now a 4-tuple of

    * the Token instance
    * the whole line as a string
    * the line number (the real one, counting continuation lines)
    * the position on the line of the end of the token.

    Original docstring ::

        The generate_tokens() generator requires one argment, readline, which
        must be a callable object which provides the same interface as the
        readline() method of built-in file objects. Each call to the function
        should return one line of input as a string.

        The generator produces 5-tuples with these members: the token type; the
        token string; a 2-tuple (srow, scol) of ints specifying the row and
        column where the token begins in the source; a 2-tuple (erow, ecol) of
        ints specifying the row and column where the token ends in the source;
        and the line on which the token was found. The line passed is the
        logical line; continuation lines are included.
    """
    token_list = []
    lnum = parenlev = continued = 0
    namechars = NAMECHARS
    numchars = NUMCHARS
    contstr, needcont = '', 0
    contline = None
    indents = [0]
    altindents = [0]
    last_comment = ''
    parenlevstart = (0, 0, "")

    # make the annotator happy
    endDFA = automata.DFA([], [])
    # make the annotator happy
    line = ''
    pos = 0
    lines.append("")
    strstart = (0, 0, "")
    for line in lines:
        lnum = lnum + 1
        pos, max = 0, len(line)

        if contstr:
            if not line:
                raise TokenError(
                    "EOF while scanning triple-quoted string literal",
                    strstart[2], strstart[0], strstart[1]+1,
                    token_list, lnum-1)
            endmatch = endDFA.recognize(line)
            if endmatch >= 0:
                pos = end = endmatch
                tok = (tokens.STRING, contstr + line[:end], strstart[0],
                       strstart[1], line)
                token_list.append(tok)
                last_comment = ''
                contstr, needcont = '', 0
                contline = None
            elif (needcont and not line.endswith('\\\n') and
                               not line.endswith('\\\r\n')):
                tok = (tokens.ERRORTOKEN, contstr + line, strstart[0],
                       strstart[1], line)
                token_list.append(tok)
                last_comment = ''
                contstr = ''
                contline = None
                continue
            else:
                contstr = contstr + line
                contline = contline + line
                continue

        elif parenlev == 0 and not continued:  # new statement
            if not line: break
            column = 0
            altcolumn = 0
            while pos < max:                   # measure leading whitespace
                if line[pos] == ' ':
                    column = column + 1
                    altcolumn = altcolumn + 1
                elif line[pos] == '\t':
                    column = (column/tabsize + 1)*tabsize
                    altcolumn = (altcolumn/alttabsize + 1)*alttabsize
                elif line[pos] == '\f':
                    column = 0
                else:
                    break
                pos = pos + 1
            if pos == max: break

            if line[pos] in '#\r\n':
                # skip comments or blank lines
                continue

            if column == indents[-1]:
                if altcolumn != altindents[-1]:
                    raise TabError(lnum, pos, line)
            elif column > indents[-1]:           # count indents or dedents
                if altcolumn <= altindents[-1]:
                    raise TabError(lnum, pos, line)
                indents.append(column)
                altindents.append(altcolumn)
                token_list.append((tokens.INDENT, line[:pos], lnum, 0, line))
                last_comment = ''
            else:
                while column < indents[-1]:
                    indents = indents[:-1]
                    altindents = altindents[:-1]
                    token_list.append((tokens.DEDENT, '', lnum, pos, line))
                    last_comment = ''
                if column != indents[-1]:
                    err = "unindent does not match any outer indentation level"
                    raise TokenIndentationError(err, line, lnum, 0, token_list)
                if altcolumn != altindents[-1]:
                    raise TabError(lnum, pos, line)

        else:                                  # continued statement
            if not line:
                if parenlev > 0:
                    lnum1, start1, line1 = parenlevstart
                    raise TokenError("parenthesis is never closed", line1,
                                     lnum1, start1 + 1, token_list, lnum)
                raise TokenError("EOF in multi-line statement", line,
                                 lnum, 0, token_list)
            continued = 0

        while pos < max:
            pseudomatch = pseudoDFA.recognize(line, pos)
            if pseudomatch >= 0:                            # scan for tokens
                # JDR: Modified
                start = whiteSpaceDFA.recognize(line, pos)
                if start < 0:
                    start = pos
                end = pseudomatch

                if start == end:
                    raise TokenError("Unknown character", line,
                                     lnum, start + 1, token_list)

                pos = end
                token, initial = line[start:end], line[start]
                if (initial in numchars or \
                   (initial == '.' and token != '.' and token != '...')):
                    # ordinary number
                    token_list.append((tokens.NUMBER, token, lnum, start, line))
                    last_comment = ''
                elif initial in '\r\n':
                    if parenlev <= 0:
                        tok = (tokens.NEWLINE, last_comment, lnum, start, line)
                        token_list.append(tok)
                    last_comment = ''
                elif initial == '#':
                    # skip comment
                    last_comment = token
                elif token in triple_quoted:
                    endDFA = endDFAs[token]
                    endmatch = endDFA.recognize(line, pos)
                    if endmatch >= 0:                     # all on one line
                        pos = endmatch
                        token = line[start:pos]
                        tok = (tokens.STRING, token, lnum, start, line)
                        token_list.append(tok)
                        last_comment = ''
                    else:
                        strstart = (lnum, start, line)
                        contstr = line[start:]
                        contline = line
                        break
                elif initial in single_quoted or \
                    token[:2] in single_quoted or \
                    token[:3] in single_quoted:
                    if token[-1] == '\n':                  # continued string
                        strstart = (lnum, start, line)
                        endDFA = (endDFAs[initial] or endDFAs[token[1]] or
                                   endDFAs[token[2]])
                        contstr, needcont = line[start:], 1
                        contline = line
                        break
                    else:                                  # ordinary string
                        tok = (tokens.STRING, token, lnum, start, line)
                        token_list.append(tok)
                        last_comment = ''
                elif (initial in namechars or              # ordinary name
                      ord(initial) >= 0x80):               # unicode identifier
                    if not verify_identifier(token):
                        raise TokenError("invalid character in identifier",
                                         line, lnum, start + 1, token_list)
                    token_list.append((tokens.NAME, token, lnum, start, line))
                    last_comment = ''
                elif initial == '\\':                      # continued stmt
                    continued = 1
                else:
                    if initial in '([{':
                        if parenlev == 0:
                            parenlevstart = (lnum, start, line)
                        parenlev = parenlev + 1
                    elif initial in ')]}':
                        parenlev = parenlev - 1
                        if parenlev < 0:
                            raise TokenError("unmatched '%s'" % initial, line,
                                             lnum, start + 1, token_list)
                    if token in python_opmap:
                        punct = python_opmap[token]
                    else:
                        punct = tokens.OP
                    token_list.append((punct, token, lnum, start, line))
                    last_comment = ''
            else:
                start = whiteSpaceDFA.recognize(line, pos)
                if start < 0:
                    start = pos
                if start<max and line[start] in single_quoted:
                    raise TokenError("EOL while scanning string literal",
                             line, lnum, start+1, token_list)
                tok = (tokens.ERRORTOKEN, line[pos], lnum, pos, line)
                token_list.append(tok)
                last_comment = ''
                pos = pos + 1

    lnum -= 1
    if not (flags & consts.PyCF_DONT_IMPLY_DEDENT):
        if token_list and token_list[-1][0] != tokens.NEWLINE:
            tok = (tokens.NEWLINE, '', lnum, 0, '\n')
            token_list.append(tok)
        for indent in indents[1:]:                # pop remaining indent levels
            token_list.append((tokens.DEDENT, '', lnum, pos, line))
    tok = (tokens.NEWLINE, '', lnum, 0, '\n')
    token_list.append(tok)

    token_list.append((tokens.ENDMARKER, '', lnum, pos, line))
    return token_list



"""
Makes a parser from a grammar source.

Inspired by Guido van Rossum's pgen2.
"""

import io
import tokenize
import token

from ayrton.parser.pyparser import parser


class PgenError(Exception):

    def __init__(self, msg, location=None):
        Exception.__init__(self, msg)
        self.location = location


class NFA(object):

    def __init__(self):
        self.arcs = []

    def arc(self, to_state, label=None):
        self.arcs.append((label, to_state))

    def find_unlabeled_states(self, into):
        if self in into:
            return
        into.add(self)
        for label, state in self.arcs:
            if label is None:
                state.find_unlabeled_states(into)


class DFA(object):

    def __init__(self, nfa_set, final_state):
        self.nfas = nfa_set
        self.is_final = final_state in nfa_set
        self.arcs = {}

    def arc(self, next, label):
        self.arcs[label] = next

    def unify_state(self, old, new):
        for label, state in self.arcs.items():
            if state is old:
                self.arcs[label] = new

    def __repr__(self):
        return "<DFA arcs=%r>" % self.arcs

    def __eq__(self, other):
        if not isinstance(other, DFA):
            # This shouldn't really happen.
            return NotImplemented
        if other.is_final != self.is_final:
            return False
        if len(self.arcs) != len(other.arcs):
            return False
        for label, state in self.arcs.items():
            try:
                other_state = other.arcs[label]
            except KeyError:
                return False
            else:
                if other_state is not state:
                    return False
        return True


def nfa_to_dfa(start, end):
    """Convert an NFA to a DFA(s)

    Each DFA is initially a set of NFA states without labels.  We start with the
    DFA for the start NFA.  Then we add labeled arcs to it pointing to another
    set of NFAs (the next state).  Finally, we do the same thing to every DFA
    that is found and return the list of states.
    """
    base_nfas = set()
    start.find_unlabeled_states(base_nfas)
    state_stack = [DFA(base_nfas, end)]
    for state in state_stack:
        arcs = {}
        for nfa in state.nfas:
            for label, sub_nfa in nfa.arcs:
                if label is not None:
                    sub_nfa.find_unlabeled_states(arcs.setdefault(label, set()))
        for label, nfa_set in arcs.items():
            for st in state_stack:
                if st.nfas == nfa_set:
                    break
            else:
                st = DFA(nfa_set, end)
                state_stack.append(st)
            state.arc(st, label)
    return state_stack

def simplify_dfa(dfa):
    changed = True
    while changed:
        changed = False
        for i, state in enumerate(dfa):
            for j in range(i + 1, len(dfa)):
                other_state = dfa[j]
                if state == other_state:
                    del dfa[j]
                    for sub_state in dfa:
                        sub_state.unify_state(other_state, state)
                    changed = True
                    break


class ParserGenerator(object):
    """NOT_RPYTHON"""

    def __init__(self, grammar_source):
        self.start_symbol = None
        self.dfas = {}
        stream = io.StringIO(grammar_source)
        self.token_stream = tokenize.generate_tokens(stream.readline)
        self.parse()
        self.first = {}
        self.add_first_sets()

    def build_grammar(self, grammar_cls):
        gram = grammar_cls()
        gram.start = self.start_symbol
        names = list(self.dfas.keys())
        names.sort()
        names.remove(self.start_symbol)
        names.insert(0, self.start_symbol)
        # First, build symbol and id mappings.
        for name in names:
            i = 256 + len(gram.symbol_ids)
            gram.symbol_ids[name] = i
            gram.symbol_names[i] = name
        # Then, iterate through again and finalize labels.
        for name in names:
            dfa = self.dfas[name]
            states = []
            for state in dfa:
                arcs = []
                for label, next in state.arcs.items():
                    arcs.append((self.make_label(gram, label), dfa.index(next)))
                states.append((arcs, state.is_final))
            gram.dfas.append((states, self.make_first(gram, name)))
            assert len(gram.dfas) - 1 == gram.symbol_ids[name] - 256
        gram.start = gram.symbol_ids[self.start_symbol]
        return gram

    def make_label(self, gram, label):
        label_index = len(gram.labels)
        if label[0].isalpha():
            # Either a symbol or a token.
            if label in gram.symbol_ids:
                if label in gram.symbol_to_label:
                    return gram.symbol_to_label[label]
                else:
                    gram.labels.append(gram.symbol_ids[label])
                    gram.symbol_to_label[label] = label_index
                    return label_index
            elif label.isupper():
                token_index = gram.TOKENS[label]
                if token_index in gram.token_ids:
                    return gram.token_ids[token_index]
                else:
                    gram.labels.append(token_index)
                    gram.token_ids[token_index] = label_index
                    return label_index
            else:
                # Probably a rule without a definition.
                raise PgenError("no such rule: %r" % (label,))
        else:
            # A keyword or operator.
            value = label.strip("\"'")
            if value[0].isalpha():
                if value in gram.keyword_ids:
                    return gram.keyword_ids[value]
                else:
                    gram.labels.append(gram.KEYWORD_TOKEN)
                    gram.keyword_ids[value] = label_index
                    return label_index
            else:
                try:
                    token_index = gram.OPERATOR_MAP[value]
                except KeyError:
                    raise PgenError("no such operator: %r" % (value,))
                if token_index in gram.token_ids:
                    return gram.token_ids[token_index]
                else:
                    gram.labels.append(token_index)
                    gram.token_ids[token_index] = label_index
                    return label_index

    def make_first(self, gram, name):
        original_firsts = self.first[name]
        firsts = dict()
        for label in original_firsts:
            firsts[self.make_label(gram, label)] = None
        return firsts

    def add_first_sets(self):
        for name, dfa in self.dfas.items():
            if name not in self.first:
                self.get_first(name, dfa)

    def get_first(self, name, dfa):
        self.first[name] = None
        state = dfa[0]
        all_labels = set()
        overlap_check = {}
        for label, sub_state in state.arcs.items():
            if label in self.dfas:
                if label in self.first:
                    new_labels = self.first[label]
                    if new_labels is None:
                        raise PgenError("recursion in rule: %r" % (name,))
                else:
                    new_labels = self.get_first(label, self.dfas[label])
                all_labels.update(new_labels)
                overlap_check[label] = new_labels
            else:
                all_labels.add(label)
                overlap_check[label] = set((label,))
        inverse = {}
        for label, their_first in overlap_check.items():
            for sub_label in their_first:
                if sub_label in inverse:
                    raise PgenError("ambiguous symbol with label %s"
                                    % (label,))
                inverse[sub_label] = label
        self.first[name] = all_labels
        return all_labels

    def expect(self, token_type, value=None):
        if token_type != self.type:
            expected = token.tok_name[token_type]
            got = token.tok_name[self.type]
            raise PgenError("expected token %s but got %s" % (expected, got),
                            self.location)
        current_value = self.value
        if value is not None:
            if value != current_value:
                msg = "expected %r but got %r" % (value, current_value)
                raise PgenError(msg,self.location)
        self.advance_token()
        return current_value

    def test_token(self, token_type, value):
        if self.type == token_type and self.value == value:
            return True
        return False

    def advance_token(self):
        data = next(self.token_stream)
        # Ignore comments and non-logical newlines.
        while data[0] in (tokenize.NL, tokenize.COMMENT):
            data = next(self.token_stream)
        self.type, self.value = data[:2]
        self.location = data[2:]

    def parse(self):
        self.advance_token()
        while self.type != token.ENDMARKER:
            # Skip over whitespace.
            while self.type == token.NEWLINE:
                self.advance_token()
            name, start_state, end_state = self.parse_rule()
            dfa = nfa_to_dfa(start_state, end_state)
            simplify_dfa(dfa)
            self.dfas[name] = dfa
            if self.start_symbol is None:
                self.start_symbol = name

    def parse_rule(self):
        # RULE: NAME ':' ALTERNATIVES
        name = self.expect(token.NAME)
        self.expect(token.OP, ":")
        start_state, end_state = self.parse_alternatives()
        self.expect(token.NEWLINE)
        return name, start_state, end_state

    def parse_alternatives(self):
        # ALTERNATIVES: ITEMS ('|' ITEMS)*
        first_state, end_state = self.parse_items()
        if self.test_token(token.OP, "|"):
            # Link all alternatives into a enclosing set of states.
            enclosing_start_state = NFA()
            enclosing_end_state = NFA()
            enclosing_start_state.arc(first_state)
            end_state.arc(enclosing_end_state)
            while self.test_token(token.OP, "|"):
                self.advance_token()
                sub_start_state, sub_end_state = self.parse_items()
                enclosing_start_state.arc(sub_start_state)
                sub_end_state.arc(enclosing_end_state)
            first_state = enclosing_start_state
            end_state = enclosing_end_state
        return first_state, end_state

    def parse_items(self):
        # ITEMS: ITEM+
        first_state, end_state = self.parse_item()
        while self.type in (token.STRING, token.NAME) or \
                           self.test_token(token.OP, "(") or \
                           self.test_token(token.OP, "["):
            sub_first_state, new_end_state = self.parse_item()
            end_state.arc(sub_first_state)
            end_state = new_end_state
        return first_state, end_state

    def parse_item(self):
        # ITEM: '[' ALTERNATIVES ']' | ATOM ['+' | '*']
        if self.test_token(token.OP, "["):
            self.advance_token()
            start_state, end_state = self.parse_alternatives()
            self.expect(token.OP, "]")
            # Bypass the rule if this is optional.
            start_state.arc(end_state)
            return start_state, end_state
        else:
            atom_state, next_state = self.parse_atom()
            # Check for a repeater.
            if self.type == token.OP and self.value in ("+", "*"):
                next_state.arc(atom_state)
                repeat = self.value
                self.advance_token()
                if repeat == "*":
                    # Optionally repeated
                    return atom_state, atom_state
                else:
                    # Required
                    return atom_state, next_state
            else:
                return atom_state, next_state

    def parse_atom(self):
        # ATOM: '(' ALTERNATIVES ')' | NAME | STRING
        if self.test_token(token.OP, "("):
            self.advance_token()
            rule = self.parse_alternatives()
            self.expect(token.OP, ")")
            return rule
        elif self.type in (token.NAME, token.STRING):
            atom_state = NFA()
            next_state = NFA()
            atom_state.arc(next_state, self.value)
            self.advance_token()
            return atom_state, next_state
        else:
            invalid = token.tok_name[self.type]
            raise PgenError("unexpected token: %s" % (invalid,),
                            self.location)



"""
Various flags used during the compilation process.
"""

CO_OPTIMIZED = 0x0001
CO_NEWLOCALS = 0x0002
CO_VARARGS = 0x0004
CO_VARKEYWORDS = 0x0008
CO_NESTED = 0x0010
CO_GENERATOR = 0x0020
CO_NOFREE = 0x0040
CO_GENERATOR_ALLOWED = 0x1000
CO_FUTURE_DIVISION = 0x2000
CO_FUTURE_ABSOLUTE_IMPORT = 0x4000
CO_FUTURE_WITH_STATEMENT = 0x8000
CO_FUTURE_PRINT_FUNCTION = 0x10000
CO_FUTURE_UNICODE_LITERALS = 0x20000
CO_FUTURE_BARRY_AS_BDFL = 0x40000
#pypy specific:
CO_KILL_DOCSTRING = 0x100000
CO_YIELD_INSIDE_TRY = 0x200000

PyCF_MASK = (CO_FUTURE_DIVISION | CO_FUTURE_ABSOLUTE_IMPORT |
             CO_FUTURE_WITH_STATEMENT | CO_FUTURE_PRINT_FUNCTION |
             CO_FUTURE_UNICODE_LITERALS | CO_FUTURE_BARRY_AS_BDFL)
PyCF_SOURCE_IS_UTF8 = 0x0100
PyCF_DONT_IMPLY_DEDENT = 0x0200
PyCF_ONLY_AST = 0x0400
PyCF_IGNORE_COOKIE = 0x0800
PyCF_ACCEPT_NULL_BYTES = 0x10000000   # PyPy only, for compile()



# -*- coding: utf-8 -*-

# (c) 2013 Marcos Dione <mdione@grulic.org.ar>

# This file is part of ayrton.
#
# ayrton is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ayrton is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ayrton.  If not, see <http://www.gnu.org/licenses/>.

import ast
from ast import Pass, Module, Bytes, copy_location, Call, Name, Load, Str, BitOr
from ast import fix_missing_locations, Import, alias, Attribute, ImportFrom
from ast import keyword, Gt, Lt, GtE, RShift, Tuple, FunctionDef, arguments
from ast import Store, Assign, Subscript
import pickle
from collections import defaultdict
import logging

logger= logging.getLogger ('ayrton.castt')

import ayrton
from ayrton.ast_pprinter import pprint

def append_to_tuple (t, o):
    l= list (t)
    l.append (o)
    return tuple (l)

def pop_from_tuple (t, n=-1):
    l= list (t)
    l.pop (n)
    return tuple (l)

def is_executable (node):
    # Call(func=Call(func=Name(id='Command', ctx=Load()),
    #                args=[Str(s='ls')], keywords=[], starargs=None, kwargs=None),
    #      args=[Str(s='-l')], keywords=[], starargs=None, kwargs=None)
    return (type (node)==Call and
            type (node.func)==Call and
            type (node.func.func)==Name and
            node.func.func.id=='Command')

def is_option (arg):
    return type (arg)==Call and type (arg.func)==Name and arg.func.id=='o'

def has_keyword (node, keyword):
    return any ([kw.arg==keyword for kw in node.keywords])

def update_keyword (node, keyword):
    found= False
    for i in range (len (node.keywords)):
        if node.keywords[i].arg==keyword.arg:
            # TODO: warn
            node.keywords[i]= keyword
            found= True
            # there can't be two
            break

    if not found:
        node.keywords.append (keyword)

def func_name2dotted_exec (node):
    logger.debug (ast.dump (node))

    complete_name= str (pprint (node))

    while type (node) in (Attribute, Subscript):
        node= node.value

    return (node.id, complete_name)

class CrazyASTTransformer (ast.NodeTransformer):
    def __init__ (self, environ, file_name=None):
        super ().__init__ ()
        # the whole ayrton instace globals
        self.environ= environ
        # names defined in the global namespace
        self.known_names= defaultdict (lambda: 0)
        # holds the names in the stack so far
        # NOTE: it must be a tuple so it can be hashable
        self.stack= ()
        # holds the temporary namespaces in function and class definitions
        # key: the stack so far
        # value: list of names
        self.defined_names= defaultdict (list)
        # for testing
        self.seen_names= set ()
        self.file_name= file_name

    def modify (self, tree):
        m= self.visit (tree)
        ast.fix_missing_locations (m)

        return m

    # The following constructs bind names:
    # [x] formal parameters to functions,
    # [x] import statements,
    # [x] class and
    # [x] function definitions (these bind the class or function name
    #     in the defining block),
    # [x] and targets that are identifiers if occurring in an assignment,
    # [x] for loop header, or
    # [x] after as in a with statement
    # [x] or except clause.
    # The import statement of the form from ... import * binds all names defined
    # in the imported module, except those beginning with an underscore.

    # A block is a piece of Python program text that is executed as a unit.
    # The following are blocks:
    # [x] a module,
    # [x] a function body, and
    # [x] a class definition.
    # [ ] A script file is a code block.
    # [ ] The string argument passed to the built-in functions eval() and exec()
    #     is a code block.

    # A scope defines the visibility of a name within a block. If a local variable
    # is defined in a block, its scope includes that block. If the definition
    # occurs in a function block, the scope extends to any blocks contained within
    # the defining one, unless a contained block introduces a different binding
    # for the name. The scope of names defined in a class block is limited to the
    # class block; it does not extend to the code blocks of methods – this
    # includes comprehensions and generator expressions since they are implemented
    # using a function scope.

    # [x] A target occurring in a del statement is also considered bound for this
    #     purpose (though the actual semantics are to unbind the name).

    def bind (self, o):
        name= None

        if type (o)==Name:
            # Name(id='a', ctx=Store())
            name= o.id

        elif type (o)==str:
            name= o

        elif type (o)==Tuple:
            # Tuple(elts=[Name(id='a', ctx=Store()), Name(id='b', ctx=Store())])
            for elt in o.elts:
                self.bind (elt)

        elif type (o)==list:
            for e in o:
                self.bind (e)

        if name is not None:
            self.known_names[name]+= 1
            self.defined_names[self.stack].append (name)
            self.seen_names.add (name)

    def unbind (self, name, remove_from_stack=False):
        self.known_names[name]-= 1
        if remove_from_stack:
            self.defined_names[self.stack].remove (name)

    def visit_Import (self, node):
        self.generic_visit (node)
        # Import(names=[alias(name='foo', asname=None)])
        for name in node.names:
            if name.asname is not None:
                n= name.asname
            else:
                # 'import os.path' -> only save 'os'
                # NOTE: why? doesn't this clash with resolving foo.py as executable name?
                n= name.name.split ('.')[0]

            logger.debug (n)
            self.bind (n)

        return node

    visit_ImportFrom= visit_Import
        # ImportFrom(module='bar', names=[alias(name='baz', asname=None)], level=0)

    def visit_ClassDef (self, node):
        # ClassDef(name='foo', bases=[], keywords=[], starargs=None, kwargs=None,
        #          body=[Pass()], decorator_list=[])
        self.bind (node.name)

        self.stack= append_to_tuple (self.stack, node.name)
        self.generic_visit (node)

        # take out the function from the stack
        names= self.defined_names[self.stack]
        self.stack= pop_from_tuple (self.stack)
        # ... and unbind the names defined in it from the known_names
        for name in names:
            self.unbind (name)

        return node

    def visit_FunctionDef (self, node):
        # FunctionDef(name='foo',
        #             args=arguments(args=[arg(arg='a', annotation=None),
        #                                  arg(arg='b', annotation=None)],
        #                            vararg='args', varargannotation=None,
        #                            kwonlyargs=[], kwarg='kwargs',
        #                            kwargannotation=None, defaults=[Num(n=1)],
        #                            kw_defaults=[]),
        #             body=[Pass()], decorator_list=[], returns=None)

        # add the name as local name
        self.bind (node.name)

        self.stack= append_to_tuple (self.stack, node.name)

        # add the arguments as local names
        for arg in node.args.args:
            self.known_names[arg.arg]+= 1
            self.defined_names[self.stack].append (arg.arg)

        self.generic_visit (node)

        # take out the function from the stack
        names= self.defined_names[self.stack]
        self.stack= pop_from_tuple (self.stack)
        # ... and remove the names defined in it from the known_names
        for name in names:
            self.unbind (name)

        return node

    visit_AsyncFunctionDef= visit_FunctionDef

    def visit_For (self, node):
        # For(target=Name(id='x', ctx=Store()), iter=List(elts=[], ctx=Load()),
        #     body=[Pass()], orelse=[])
        # For(target=Tuple(elts=[Name(id='band', ctx=Store()), Name(id='color', ctx=Store())], ctx=Store()),
        #     iter=Tuple(elts=[...], ctx=Load()),
        #     body=[Pass()], orelse=[])
        self.generic_visit (node)
        self.bind (node.target)

        # if iter is Command, _out=Capture
        # so this works as expected:
        # for line in ls(): ...

        # For(target=Name(id='line', ctx=Store()),
        #     iter=Call(func=Call(func=Name(id='Command', ctx=Load()), ...
        if is_executable (node.iter):
            update_keyword (node.iter,
                            keyword (arg='_out', value=Name (id='Capture', ctx=Load ())))

        return node

    visit_AsyncFor= visit_For

    def visit_ExceptHandler (self, node):
        # ExceptHandler(type=Name(id='A', ctx=Load()),
        #               name='e',
        #               body=[Pass()])
        self.bind (node.name)
        self.generic_visit (node)

        return node

    def visit_Assign (self, node):
        # Assign(targets=[Tuple(elts=[Name(id='a', ctx=Store()), Name(id='b', ctx=Store())], ctx=Store())],
        #        value=Tuple(elts=[Num(n=4), Num(n=2)], ctx=Load()))
        self.generic_visit (node)
        self.bind (node.targets)

        return node

    def visit_Delete (self, node):
        self.generic_visit (node)
        # Delete(targets=[Name(id='foo', ctx=Del())])
        for name in node.targets:
            self.unbind (name.id, remove_from_stack=True)

        return node

    # BinOp(left=BinOp(left=Name(id='a', ctx=Load()),
    #                  op=BitOr(), right=Name(id='b', ctx=Load())),
    #       op=BitOr(), right=Name(id='c', ctx=Load()))
    def visit_BinOp (self, node):
        self.generic_visit (node)

        # BinOp( left=Call(...), op=BitOr(), right=Call(...))
        if type (node.op)==BitOr:
            # pipe
            # BinOp (left, BitOr, right) -> right (left, ...)

            # check the left and right; if they're calls to Command
            # then do the magic

            # NOTE: they're Commands only because the children have already been
            # visited and transformed to such things

            both= is_executable (node.left) and is_executable (node.right)

            if both:
                # left (...) | right (...)

                # r, w= os.pipe ()
                # left (..., _out=w)
                # os.close (w)
                # right (..., _in=r)
                # os.close (r)

                # Assign(targets=[Tuple(elts=[Name(id='r', ctx=Store()),
                #                             Name(id='w', ctx=Store())], ctx=Store())],
                #        value=Call(func=Attribute(value=Name(id='os', ctx=Load()),
                #                                  attr='pipe', ctx=Load()),
                #                   args=[], keywords=[], starargs=None, kwargs=None)),
                # Expr(value=Call(func=Name(id='echo', ctx=Load()),
                #                 args=[Str(s='pipe!')],
                #                 keywords=[keyword(arg='_out', value=Name(id='w', ctx=Load()))],
                #                 starargs=None, kwargs=None)),
                # Expr(value=Call(func=Attribute(value=Name(id='os', ctx=Load()), attr='close', ctx=Load()),
                #                 args=[Name(id='w', ctx=Load())],
                #                 keywords=[], starargs=None, kwargs=None)),
                # Expr(value=Call(func=Name(id='grep', ctx=Load()),
                #                 args=[Str(s='pipe')],
                #                 keywords=[keyword(arg='_in', value=Name(id='r', ctx=Load()))],
                #                 starargs=None, kwargs=None)),
                # Expr(value=Call(func=Attribute(value=Name(id='os', ctx=Load()), attr='close', ctx=Load()),
                #                 args=[Name(id='r', ctx=Load())],
                #                 keywords=[], starargs=None, kwargs=None))

                # I can't believe it's this easy
                # TODO: check if _err is not being captured instead
                update_keyword (node.left,
                                keyword (arg='_out', value=Name (id='Pipe', ctx=Load ())))
                update_keyword (node.left,
                                keyword (arg='_bg',  value=Name (id='True', ctx=Load ())))
                ast.fix_missing_locations (node.left)
                update_keyword (node.right, keyword (arg='_in', value=node.left))
                node= node.right

                # Call(func=Call(func=Attribute(value=Name(id='CommandWrapper', ctx=Load()),
                #                               attr='_create', ctx=Load()),
                #                args=[Str(s='grep')], keywords=[], starargs=None, kwargs=None),
                #      args=[Call(func=Call(func=Attribute(value=Name(id='CommandWrapper', ctx=Load()),
                #                                          attr='_create', ctx=Load()),
                #                      args=[Str(s='ls')], keywords=[], starargs=None, kwargs=None),
                #                 args=[], keywords=[keyword(arg='_out',
                #                                            value=Name(id='Capture', ctx=Load()))],
                #                 starargs=None, kwargs=None),
                #            Str(s='foo')],
                #      keywords=[], starargs=None, kwargs=None)

        elif type (node.op)==RShift:
            # BinOp(left=Call(func=Name(id='ls', ctx=Load()), args=[], keywords=[], starargs=None, kwargs=None),
            #       op=RShift(),
            #       right=Str(s='foo.txt'))
            if is_executable (node.left):
                update_keyword (node.left,
                                keyword (arg='_out',
                                         value=Call (func=Name (id='open', ctx=Load ()),
                                                     args=[node.right, Str (s='ab')],
                                                     keywords=[], starargs=None,
                                                     kwargs=None)))
                ast.fix_missing_locations (node.left)
                node= node.left

        return node

    # Compare(left=Call(func=Name(id='grep', ctx=Load()), args=[Str(s='root')], keywords=[], starargs=None, kwargs=None),
    #        ops=[Lt()],
    #        comparators=[BinOp(left=Str(s='/etc/passwd'),
    #                           op=RShift(),
    #                           right=Str(s='/tmp/foo'))])
    def visit_Compare (self, node):
        self.generic_visit (node)
        ans= node

        # ls () < "bar.txt" > "foo.txt"
        # Compare(left=Call(func=Name(id='ls', ctx=Load()), args=[], keywords=[],
        #                        starargs=None, kwargs=None),
        #         ops=[Lt(), Gt()],
        #         comparators=[Str(s='bar.txt'), Str(s='foo.txt')]))

        if is_executable (node.left):
            # yes, they're reversed, but it makes more sense to me like this
            for comp, op in zip (node.ops, node.comparators):
                if type (comp)==Gt:
                    # > means _out
                    update_keyword (node.left, keyword (arg='_out', value=op))
                    ans= node.left
                elif type (comp)==GtE:
                    # >= means _out+_err
                    update_keyword (node.left, keyword (arg='_out', value=op))
                    update_keyword (node.left, keyword (arg='_err_to_out', value=op))
                    ans= node.left
                elif type (comp)==Lt:
                    # < means _in
                    update_keyword (node.left, keyword (arg='_in', value=op))
                    ans= node.left

        return ans

    def visit_Call (self, node):
        self.generic_visit (node)
        # Call(func=Name(id='b', ctx=Load()), args=[], keywords=[], starargs=None,
        #      kwargs=None)
        # Call(func=Attribute(value=Name(id='test', ctx=Load()), attr='py', ctx=Load()), ...)
        # NOTE: what other things can be the func part?
        name= None
        unknown= False

        if type (node.func)==Name:
            name= func_name= node.func.id
            defs= self.known_names[func_name]
            if defs==0:
                unknown= True

        elif type (node.func)==Attribute:
            name, func_name= func_name2dotted_exec (node.func)
            defs= self.known_names[name]
            if defs==0:
                unknown= True

        logger.debug2 ("%s: %s", name, unknown)

        if unknown:
            if not name in self.environ:
                # it's not one of the builtin functions
                # I guess I have no other option but to try to execute
                # something here...
                new_node= Call (func=Name (id='Command', ctx=Load ()),
                                args=[Str (s=func_name)], keywords=[],
                                starargs=None, kwargs=None)

                # check if the first parameter is a Command; if so, redirect
                # its output, remove it from the args and put it in the _in
                # kwarg
                if len (node.args)>0:
                    first_arg= node.args[0]
                    if is_executable (first_arg) and not has_keyword (first_arg, '_err'):
                        out= keyword (arg='_out', value=Name (id='Pipe', ctx=Load ()))
                        update_keyword (first_arg, out)
                        bg= keyword (arg='_bg', value=Name (id='True', ctx=Load ()))
                        update_keyword (first_arg, bg)
                        node.args.pop (0)
                        update_keyword (node, keyword (arg='_in', value=first_arg))

                    for arg in node.args:
                        if is_option (arg):
                            # ast_pprinter takes care of expressions
                            kw= arg.keywords[0]
                            logger.debug2 ("->>>kw: %s", ast.dump (kw))
                            kw.arg= str (pprint (kw.arg))

                ast.copy_location (new_node, node)
                node.func= new_node
                ast.fix_missing_locations (node)
            else:
                # this is a normal function call
                # Call(func=Name(id='foo', ctx=Load()),
                #      args=[Call(func=Name(id='o', ctx=Load()),
                #                 args=[], keywords=[keyword(arg='a', value=Num(n=42))],
                #                 starargs=None, kwargs=None)],
                #      keywords=[], starargs=None, kwargs=None))
                # the parser has converted all the keyword arguments to o(k=v)
                # we need to convert them back to keywords and apply Python's syntax rules
                new_args= []
                used_keywords= set ()
                first_kw= False
                for index, arg in enumerate (node.args):
                    # NOTE: maybe o() can be left in its own namespace so it doesn't pollute
                    if is_option (arg):
                        kw_expr= arg.keywords[0].arg
                        if not isinstance (kw_expr, ast.Name) and not isinstance (kw_expr, str):
                            raise SyntaxError (self.file_name, node.lineno, node.col_offset,
                                               "keyword can't be an expression")

                        if isinstance (kw_expr, ast.Name):
                            kw_name= kw_expr.id
                        else:
                            kw_name= kw_expr  # str

                        if kw_name in used_keywords:
                            raise SyntaxError (self.file_name, node.lineno, node.col_offset,
                                               "keyword argument repeated")

                        # convert the expr into a str
                        new_kw= keyword (kw_name, arg.keywords[0].value)
                        node.keywords.append (new_kw)
                        used_keywords.add (kw_name)
                        first_kw= True
                    else:
                        if first_kw:
                            raise SyntaxError (self.file_name, node.lineno, node.col_offset,
                                               "non-keyword arg after keyword arg")

                        new_args.append (arg)

                node.args= new_args

        return node

    def visit_With (self, node):
        # With(items=[withitem(context_expr=Call(func=Name(id='foo', ctx=Load()),
        #                                        args=[], keywords=[], starargs=None,
        #                                        kwargs=None),
        #                      optional_vars=Name(id='bar', ctx=Store()))], body=[Pass()])
        for item in node.items:
            if item.optional_vars is not None:
                self.bind (item.optional_vars.id)

        self.generic_visit (node)

        # handle 'remote'
        sub_node= node.items[0].context_expr
        if (type (sub_node)==Call and hasattr (sub_node.func, 'id') and
            sub_node.func.id=='remote'):
            # capture the body and put it as the first argument to ssh()
            # but within a module, and already pickled;
            # otherwise we need to create an AST for the call of all the
            # constructors in the body we capture... it's complicated
            m= Module (body=node.body)

            data= pickle.dumps (m)
            s= Bytes (s=data)
            s.lineno= node.lineno
            s.col_offset= node.col_offset
            sub_node.args.insert (0, s)

            p= Pass ()
            p.lineno= node.lineno+1
            p.col_offset= node.col_offset+4

            node.body= [ p ]

        return node

    visit_AsyncWith= visit_With



"""Python token definitions."""

python_tokens = {}
python_opmap = {}

def _add_tok(name, *values):
    index = len(python_tokens)
    assert index < 256
    python_tokens[name] = index
    for value in values:
        python_opmap[value] = index

_add_tok('ENDMARKER')
_add_tok('NAME')
_add_tok('NUMBER')
_add_tok('STRING')
_add_tok('NEWLINE')
_add_tok('INDENT')
_add_tok('DEDENT')
_add_tok('LPAR', "(")
_add_tok('RPAR', ")")
_add_tok('LSQB', "[")
_add_tok('RSQB', "]")
_add_tok('COLON', ":")
_add_tok('COMMA',  "," )
_add_tok('SEMI', ";" )
_add_tok('PLUS', "+" )
_add_tok('MINUS', "-" )
_add_tok('STAR', "*" )
_add_tok('SLASH', "/" )
_add_tok('VBAR', "|" )
_add_tok('AMPER', "&" )
_add_tok('LESS', "<" )
_add_tok('GREATER', ">" )
_add_tok('EQUAL', "=" )
_add_tok('DOT', "." )
_add_tok('PERCENT', "%" )
_add_tok('BACKQUOTE', "`" )
_add_tok('LBRACE', "{" )
_add_tok('RBRACE', "}" )
_add_tok('EQEQUAL', "==" )
_add_tok('NOTEQUAL', "!=", "<>" )
_add_tok('LESSEQUAL', "<=" )
_add_tok('GREATEREQUAL', ">=" )
_add_tok('TILDE', "~" )
_add_tok('CIRCUMFLEX', "^" )
_add_tok('LEFTSHIFT', "<<" )
_add_tok('RIGHTSHIFT', ">>" )
_add_tok('DOUBLESTAR', "**" )
_add_tok('PLUSEQUAL', "+=" )
_add_tok('MINEQUAL', "-=" )
_add_tok('STAREQUAL', "*=" )
_add_tok('SLASHEQUAL', "/=" )
_add_tok('PERCENTEQUAL', "%=" )
_add_tok('AMPEREQUAL', "&=" )
_add_tok('VBAREQUAL', "|=" )
_add_tok('CIRCUMFLEXEQUAL', "^=" )
_add_tok('LEFTSHIFTEQUAL', "<<=" )
_add_tok('RIGHTSHIFTEQUAL', ">>=" )
_add_tok('DOUBLESTAREQUAL', "**=" )
_add_tok('DOUBLESLASH', "//" )
_add_tok('DOUBLESLASHEQUAL',"//=" )
_add_tok('AT', "@" )
_add_tok('RARROW', "->")
_add_tok('ELLIPSIS', "...")
_add_tok('OP')
_add_tok('ERRORTOKEN')

# extra PyPy-specific tokens
_add_tok("COMMENT")
_add_tok("NL")

del _add_tok




def parse_future(tree, feature_flags):
    from pypy.interpreter.astcompiler import ast
    future_lineno = 0
    future_column = 0
    flags = 0
    have_docstring = False
    body = None
    if isinstance(tree, ast.Module):
        body = tree.body
    elif isinstance(tree, ast.Interactive):
        body = tree.body
    if body is None:
        return 0, 0, 0
    for stmt in body:
        if isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Str):
            if have_docstring:
                break
            else:
                have_docstring = True
        elif isinstance(stmt, ast.ImportFrom):
            if stmt.module == "__future__":
                future_lineno = stmt.lineno
                future_column = stmt.col_offset
                for alias in stmt.names:
                    assert isinstance(alias, ast.alias)
                    # If this is an invalid flag, it will be caught later in
                    # codegen.py.
                    flags |= feature_flags.get(alias.name, 0)
            else:
                break
        else:
            break
    return flags, future_lineno, future_column


class ForbiddenNameAssignment(Exception):

    def __init__(self, name, node):
        self.name = name
        self.node = node


def check_forbidden_name(name, node=None):
    """Raise an error if the name cannot be assigned to."""
    if name in ("None", "__debug__"):
        raise ForbiddenNameAssignment(name, node)
    # XXX Warn about using True and False


# shamelessly lifted off rpython
class unrolling_iterable:

    def __init__(self, iterable):
        self._items = list(iterable)
        self._head = _unroller(self._items)

    def __iter__(self):
        return iter(self._items)

    def get_unroller(self):
        return self._head


# ditto
class _unroller:

    def __init__(self, items, i=0):
        self._items = items
        self._i = i
        self._next = None

    def step(self):
        v = self._items[self._i]
        if self._next is None:
            self._next = _unroller(self._items, self._i+1)
        return v, self._next


def dict_to_switch(d):
    """Convert of dictionary with integer keys to a switch statement."""
    def lookup(query):
        return d[query]
    lookup._always_inline_ = True
    unrolling_items = unrolling_iterable(d.items())
    return lookup


def mangle(name, klass):
    if not name.startswith('__'):
        return name
    # Don't mangle __id__ or names with dots. The only time a name with a dot
    # can occur is when we are compiling an import statement that has a package
    # name.
    if name.endswith('__') or '.' in name:
        return name
    try:
        i = 0
        while klass[i] == '_':
            i = i + 1
    except IndexError:
        return name
    return "_%s%s" % (klass[i:], name)


def intern_if_common_string(space, w_const):
    # only intern identifier-like strings
    from pypy.objspace.std.unicodeobject import _isidentifier
    if (space.is_w(space.type(w_const), space.w_unicode) and
        _isidentifier(space.unicode_w(w_const))):
        return space.new_interned_w_str(w_const)
    return w_const


def new_identifier(space, name):
    # Check whether there are non-ASCII characters in the identifier; if
    # so, normalize to NFKC
    for c in name:
        if ord(c) > 0x80:
            break
    else:
        return name

    from pypy.module.unicodedata.interp_ucd import ucd
    w_name = space.wrap(name.decode('utf-8'))
    w_id = space.call_method(ucd, 'normalize', space.wrap('NFKC'), w_name)
    return space.unicode_w(w_id).encode('utf-8')



import __future__ as future

class FutureFlags(object):

    def __init__(self, version):
        compiler_flags = 0
        self.compiler_features = {}
        self.mandatory_flags = 0
        for fname in future.all_feature_names:
            feature = getattr(future, fname)
            if version >= feature.getOptionalRelease():
                flag = feature.compiler_flag
                compiler_flags |= flag
                self.compiler_features[fname] = flag
            if version >= feature.getMandatoryRelease():
                self.mandatory_flags |= feature.compiler_flag
        self.allowed_flags = compiler_flags

    def get_flag_names(self, space, flags):
        flag_names = []
        for name, value in self.compiler_features.items():
            if flags & value:
                flag_names.append(name)
        return flag_names

    def get_compiler_feature(self, name):
        return self.compiler_features.get(name, 0)

futureFlags_2_4 = FutureFlags((2, 4, 4, 'final', 0))
futureFlags_2_5 = FutureFlags((2, 5, 0, 'final', 0))
futureFlags_2_7 = FutureFlags((2, 7, 0, 'final', 0))
futureFlags_3_2 = FutureFlags((3, 2, 0, 'final', 0))


class TokenIterator:
    def __init__(self, tokens):
        self.tokens = tokens
        self.index = 0
        self.next()

    def next(self):
        index = self.index
        self.index = index + 1
        self.tok = self.tokens[index]

    def skip(self, n):
        if self.tok[0] == n:
            self.next()
            return True
        else:
            return False

    def skip_name(self, name):
        from ayrton.parser.pyparser import pygram
        if self.tok[0] == pygram.tokens.NAME and self.tok[1] == name:
            self.next()
            return True
        else:
            return False

    def next_feature_name(self):
        from ayrton.parser.pyparser import pygram
        if self.tok[0] == pygram.tokens.NAME:
            name = self.tok[1]
            self.next()
            if self.skip_name("as"):
                self.skip(pygram.tokens.NAME)
            return name
        else:
            return ''

    def skip_newlines(self):
        from ayrton.parser.pyparser import pygram
        while self.skip(pygram.tokens.NEWLINE):
            pass


def add_future_flags(future_flags, tokens):
    from ayrton.parser.pyparser import pygram
    it = TokenIterator(tokens)
    result = 0
    #
    # The only things that can precede a future statement are another
    # future statement and a doc string (only one).  This is a very
    # permissive parsing of the given list of tokens; it relies on
    # the real parsing done afterwards to give errors.
    it.skip_newlines()
    it.skip_name("r") or it.skip_name("u") or it.skip_name("ru")
    if it.skip(pygram.tokens.STRING):
        it.skip_newlines()

    while (it.skip_name("from") and
           it.skip_name("__future__") and
           it.skip_name("import")):
        it.skip(pygram.tokens.LPAR)    # optionally
        result |= future_flags.get_compiler_feature(it.next_feature_name())
        while it.skip(pygram.tokens.COMMA):
            result |= future_flags.get_compiler_feature(it.next_feature_name())
        it.skip(pygram.tokens.RPAR)    # optionally
        it.skip(pygram.tokens.SEMI)    # optionally
        it.skip_newlines()

    position = (it.tok[2], it.tok[3])
    return result, position



# (c) 2013 Marcos Dione <mdione@grulic.org.ar>

# This file is part of ayrton.
#
# ayrton is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ayrton is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ayrton.  If not, see <http://www.gnu.org/licenses/>.

import os
import stat

# these functions imitate the -* tests from [ (as per bash's man page)

def simple_stat (fname):
    try:
        return os.stat (fname)
    except (IOError, OSError):
        return None

def _a (fname):
    return simple_stat (fname) is not None

def _b (fname):
    s= simple_stat (fname)
    return s is not None and stat.S_ISBLK (s.st_mode)

def _c (fname):
    s= simple_stat (fname)
    return s is not None and stat.S_ISCHR (s.st_mode)

def _d (fname):
    s= simple_stat (fname)
    return s is not None and stat.S_ISDIR (s.st_mode)

# both return the same thing!
_e= _a

def _f (fname):
    s= simple_stat (fname)
    return s is not None and stat.S_ISREG (s.st_mode)

def _g (fname):
    s= simple_stat (fname)
    return s is not None and (stat.S_IMODE (s.st_mode) & stat.S_ISGID)!=0

def _h (fname):
    s= simple_stat (fname)
    return s is not None and stat.S_ISLNK (s.st_mode)

def _k (fname):
    s= simple_stat (fname)
    # VTX?!?
    return s is not None and (stat.S_IMODE (s.st_mode) & stat.S_ISVTX)!=0

def _p (fname):
    s= simple_stat (fname)
    return s is not None and stat.S_ISFIFO (s.st_mode)

def _r (fname):
    s= simple_stat (fname)
    return s is not None and (stat.S_IMODE (s.st_mode) & (stat.S_IRUSR|stat.S_IRGRP|stat.S_IROTH))!=0

def _s (fname):
    s= simple_stat (fname)
    return s is not None and s.st_size>0

# TODO: _t
# os.isatty(fd)

def _u (fname):
    s= simple_stat (fname)
    return s is not None and (stat.S_IMODE (s.st_mode) & stat.S_ISUID)!=0

def _w (fname):
    s= simple_stat (fname)
    return s is not None and (stat.S_IMODE (s.st_mode) & (stat.S_IWUSR|stat.S_IWGRP|stat.S_IWOTH))!=0

def _x (fname):
    s= simple_stat (fname)
    return s is not None and (stat.S_IMODE (s.st_mode) & (stat.S_IXUSR|stat.S_IXGRP|stat.S_IXOTH))!=0

# TODO: _G, _O

_L= _h

def _N (fname):
    s= simple_stat (fname)
    return s is not None and s.st_mtime_ns > s.st_atime_ns

def _S (fname):
    s= simple_stat (fname)
    return s is not None and stat.S_ISSOCK (s.st_mode)

# TODO: _ef,

def _nt (a, b):
    # file1 is newer (according to modification date) than file2, or if file1 exists and file2 does not.
    s1= simple_stat (a)
    s2= simple_stat (b)
    return (s1 is not None and s2 is None) or (s1.st_mtime_ns > s2.st_mtime_ns)

def _ot (a, b):
    # file1 is older than file2, or if file2 exists and file1 does not.
    s1= simple_stat (a)
    s2= simple_stat (b)
    return (s2 is not None and s1 is None) or (s2.st_mtime_ns > s1.st_mtime_ns)



"""
A CPython inspired RPython parser.
"""


class Grammar(object):
    """
    Base Grammar object.

    Pass this to ParserGenerator.build_grammar to fill it with useful values for
    the Parser.
    """

    def __init__(self):
        self.symbol_ids = {}
        self.symbol_names = {}
        self.symbol_to_label = {}
        self.keyword_ids = {}
        self.dfas = []
        self.labels = [0]
        self.token_ids = {}
        self.start = -1

    def shared_copy(self):
        new = self.__class__()
        new.symbol_ids = self.symbol_ids
        new.symbols_names = self.symbol_names
        new.keyword_ids = self.keyword_ids
        new.dfas = self.dfas
        new.labels = self.labels
        new.token_ids = self.token_ids
        return new

    def _freeze_(self):
        # Remove some attributes not used in parsing.
        try:
            del self.symbol_to_label
            del self.symbol_names
            del self.symbol_ids
        except AttributeError:
            pass
        return True


class Node(object):

    __slots__ = "type value children lineno col_offset".split()

    def __init__(self, type, value, children, lineno, column):
        self.type = type
        self.value = value
        self.children = children
        self.lineno = lineno
        self.col_offset = column

    def __eq__(self, other):
        # For tests.
        return (self.type == other.type and
                self.value == other.value and
                self.children == other.children)

    def __repr__(self):
        if self.value is None:
            return "Node(type=%s, children=%r)" % (self.type, self.children)
        else:
            return "Node(type=%s, value=%r)" % (self.type, self.value)


class ParseError(Exception):

    def __init__(self, msg, token_type, value, lineno, column, line,
                 expected=-1):
        self.msg = msg
        self.token_type = token_type
        self.value = value
        self.lineno = lineno
        self.col_offset = column
        self.line = line
        self.expected = expected

    def __str__(self):
        return "ParserError(%s, %r)" % (self.token_type, self.value)


class Parser(object):

    def __init__(self, grammar):
        self.grammar = grammar
        self.root = None
        self.stack = None

    def prepare(self, start=-1):
        """Setup the parser for parsing.

        Takes the starting symbol as an argument.
        """
        if start == -1:
            start = self.grammar.start
        self.root = None
        current_node = Node(start, None, [], 0, 0)
        self.stack = []
        self.stack.append((self.grammar.dfas[start - 256], 0, current_node))

    def add_token(self, token_type, value, lineno, column, line):
        label_index = self.classify(token_type, value, lineno, column, line)
        sym_id = 0 # for the annotator
        while True:
            dfa, state_index, node = self.stack[-1]
            states, first = dfa
            arcs, is_accepting = states[state_index]
            for i, next_state in arcs:
                sym_id = self.grammar.labels[i]
                if label_index == i:
                    # We matched a non-terminal.
                    self.shift(next_state, token_type, value, lineno, column)
                    state = states[next_state]
                    # While the only possible action is to accept, pop nodes off
                    # the stack.
                    while state[1] and not state[0]:
                        self.pop()
                        if not self.stack:
                            # Parsing is done.
                            return True
                        dfa, state_index, node = self.stack[-1]
                        state = dfa[0][state_index]
                    return False
                elif sym_id >= 256:
                    sub_node_dfa = self.grammar.dfas[sym_id - 256]
                    # Check if this token can start a child node.
                    if label_index in sub_node_dfa[1]:
                        self.push(sub_node_dfa, next_state, sym_id, lineno,
                                  column)
                        break
            else:
                # We failed to find any arcs to another state, so unless this
                # state is accepting, it's invalid input.
                if is_accepting:
                    self.pop()
                    if not self.stack:
                        raise ParseError("too much input", token_type, value,
                                         lineno, column, line)
                else:
                    # If only one possible input would satisfy, attach it to the
                    # error.
                    if len(arcs) == 1:
                        expected = sym_id
                    else:
                        expected = -1
                    raise ParseError("bad input", token_type, value, lineno,
                                     column, line, expected)

    def classify(self, token_type, value, lineno, column, line):
        """Find the label for a token."""
        if token_type == self.grammar.KEYWORD_TOKEN:
            label_index = self.grammar.keyword_ids.get(value, -1)
            if label_index != -1:
                return label_index
        label_index = self.grammar.token_ids.get(token_type, -1)
        if label_index == -1:
            raise ParseError("invalid token", token_type, value, lineno, column,
                             line)
        return label_index

    def shift(self, next_state, token_type, value, lineno, column):
        """Shift a non-terminal and prepare for the next state."""
        dfa, state, node = self.stack[-1]
        new_node = Node(token_type, value, None, lineno, column)
        node.children.append(new_node)
        self.stack[-1] = (dfa, next_state, node)

    def push(self, next_dfa, next_state, node_type, lineno, column):
        """Push a terminal and adjust the current state."""
        dfa, state, node = self.stack[-1]
        new_node = Node(node_type, None, [], lineno, column)
        self.stack[-1] = (dfa, next_state, node)
        self.stack.append((next_dfa, 0, new_node))

    def pop(self):
        """Pop an entry off the stack and make its node a child of the last."""
        dfa, state, node = self.stack.pop()
        if self.stack:
            self.stack[-1][2].children.append(node)
        else:
            self.root = node



# -*- coding: utf-8 -*-

# (c) 2013 Marcos Dione <mdione@grulic.org.ar>

# This file is part of ayrton.
#
# ayrton is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ayrton is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ayrton.  If not, see <http://www.gnu.org/licenses/>.

import ayrton
import paramiko
# from ayrton.expansion import bash
import pickle
import types
from socket import socket
from threading import Thread
import sys
import errno
import ctypes
import os
import traceback
from select import select
import io
from termios import tcgetattr, tcsetattr, TCSADRAIN
from termios import IGNPAR, ISTRIP, INLCR, IGNCR, ICRNL, IXON, IXANY, IXOFF
from termios import ISIG, ICANON, ECHO, ECHOE, ECHOK, ECHONL, IEXTEN, OPOST, VMIN, VTIME
import shutil

import logging
logger= logging.getLogger ('ayrton.remote')

class ShutUpPolicy (paramiko.MissingHostKeyPolicy):
    def missing_host_key (self, *args, **kwargs):
        pass


class InteractiveThread (Thread):
    def __init__ (self, pairs):
        super ().__init__ ()
        # so I can close them at will
        logger.debug ('%s: %s', self, pairs)
        self.pairs= pairs
        self.copy_to= dict (pairs)
        self.finished= os.pipe ()

        # we're using a tty, change all the local settings for stdin
        # directly taken from openssh (sshtty.c)
        self.orig_terminfo= tcgetattr (pairs[0][0])
        # input, output, control, local, speeds, special chars
        iflag, oflag, cflag, lflag, ispeed, ospeed, cc= self.orig_terminfo

        # turn on:
        # Ignore framing errors and parity errors
        iflag|= IGNPAR
        # turn off:
        # Strip off eighth bit
        # Translate NL to CR on input
        # Ignore carriage return on input
        # XON/XOFF flow control on output
        # (XSI) Typing any character will restart stopped output. NOTE: not needed?
        # XON/XOFF flow control on input
        iflag&= ~( ISTRIP | INLCR | IGNCR | ICRNL | IXON | IXANY | IXOFF )

        # turn off:
        # When any of the characters INTR, QUIT, SUSP, or DSUSP are received, generate the corresponding signal
        # canonical mode
        # Echo input characters (finally)
        # NOTE: why these three? they only work with ICANON and we're disabling it
        # If ICANON is also set, the ERASE character erases the preceding input character, and WERASE erases the preceding word
        # If ICANON is also set, the KILL character erases the current line
        # If ICANON is also set, echo the NL character even if ECHO is not set
        # implementation-defined input processing
        lflag&= ~( ISIG | ICANON | ECHO | ECHOE | ECHOK | ECHONL | IEXTEN )

        # turn off:
        # implementation-defined output processing
        oflag&= ~OPOST

        # NOTE: whatever
        # Minimum number of characters for noncanonical read
        cc[VMIN]= 1
        # Timeout in deciseconds for noncanonical read
        cc[VTIME]= 0

        tcsetattr(self.pairs[0][0], TCSADRAIN, [ iflag, oflag, cflag, lflag,
                                                 ispeed, ospeed, cc ])

    def read (self, src):
        if isinstance (src, io.IOBase):
            data= src.read (10240)
        elif isinstance (src, int):
            data= os.read (src, 10240)
        else:
            data= src.recv (10240)

        return data


    def write (self, dst, data):
        if isinstance (dst, io.IOBase):
            dst.write (data.decode ())
            dst.flush ()
        elif isinstance (dst, int):
            os.write (dst, data)
        else:
            dst.send (data)

    def fileno (self, f):
        if isinstance (f, int):
            return f
        elif getattr (f, 'fileno', None) is not None:
            return f.fileno ()
        else:
            raise TypeError (f)

    def run (self):
        # NOTE:
        # os.sendfile (self.dst, self.src, None, 0)
        # OSError: [Errno 22] Invalid argument
        # and splice() is not available
        # so, copy by hand
        while True:
            wait_for= list (self.copy_to.keys ())
            wait_for.append (self.finished[0])
            logger.debug (wait_for)
            for wait in wait_for:
                if ( not isinstance (wait, int) and
                     (getattr (wait, 'fileno', None) is None  or
                      not isinstance (wait.fileno(), int)) ):
                    logger.debug ('type mismatch: %s', wait)

            logger.debug (wait_for)
            r, w, e= select (wait_for, [], [])

            if self.finished[0] in r:
                self.close_file (self.finished[0])
                break

            for error in e:
                logger.debug ('%s error')
                # TODO: what?

            for i in r:
                o= self.copy_to[i]
                try:
                    data= self.read (i)
                    logger.debug ('%s -> %s: %s', i, o, data)

                # ValueError: read of closed file
                except (OSError, ValueError) as e:
                    logger.debug ('stopping copying for %s', i)
                    del self.copy_to[i]
                    logger.debug (traceback.format_exc ())
                    break
                else:
                    if len (data)==0:
                        logger.debug ('stopping copying for %s, no more data', i)
                        del self.copy_to[i]
                    else:
                        self.write (o, data)

        self.close ()
        logger.debug ('%s shutdown', self)


    def close (self):
        # reset term settings
        tcsetattr (self.pairs[0][0], TCSADRAIN, self.orig_terminfo)

        for k, v in self.pairs:
            for f in (k, v):
                if ( isinstance (f, paramiko.Channel) or
                     isinstance (f, io.TextIOWrapper) ):
                     self.close_file (f)

        self.close_file (self.finished[1])


    def close_file (self, f):
        logger.debug ('closing %s', f)
        try:
            try:
                f.close ()
            except AttributeError:
                # AttributeError: 'int' object has no attribute 'close'
                os.close (f)
        except OSError as e:
            logger.debug ('closing gave %s', e)
            if e.errno!=errno.EBADF:
                raise


class RemoteStub:
    def __init__ (self, pairs):
        self.interactive= InteractiveThread (pairs)
        self.interactive.start ()


    def close (self):
        self.interactive.close ()
        self.interactive.join ()


class remote:
    "Uses the same arguments as paramiko.SSHClient.connect ()"
    def __init__ (self, ast, hostname, *args, **kwargs):
        # actually, it's not a proper ast, it's the pickle of such thing
        self.ast= ast
        self.hostname= hostname
        self.args= args

        self.param ('_debug', kwargs)
        self.param ('_test', kwargs)
        self.kwargs= kwargs

        # socket/transport where we wait for connection for the result
        self.result_listen= None
        # socket/transport where the result is going to come back
        self.result_channel= None

        self.remote= None


    def param (self, param, kwargs, default_value=False):
        """gets a param from kwargs, or uses a default_value. if found, it's
        removed from kwargs"""
        if param in kwargs:
            value= kwargs[param]
            del kwargs[param]
        else:
            value= default_value
        setattr (self, param, value)


    def __enter__ (self):
        # get the globals from the runtime

        # for solving the import problem:
        # _pickle.PicklingError: Can't pickle <class 'module'>: attribute lookup builtins.module failed
        # there are two solutions. either we setup a complex system that intercepts
        # the imports and hold them in another ayrton.Environment attribute
        # or we just weed them out here. so far this is the simpler option
        # but forces the user to reimport what's going to be used in the remote
        g= dict ([ (k, v) for k, v in ayrton.runner.globals.items ()
                          if type (v)!=types.ModuleType and k not in ('stdin', 'stdout', 'stderr') ])

        # get the locals from the runtime
        # this is not so easy: for some reason, ayrton.runner.locals is not up to
        # date in the middle of the execution (remember *this* code is executed
        # via exec() in Ayrton.run_code())
        # another option is to go through the frames
        inception_locals= sys._getframe().f_back.f_locals

        l= dict ([ (k, v) for (k, v) in inception_locals.items ()
                          if type (v)!=types.ModuleType and k not in ('stdin', 'stdout', 'stderr') ])

        # special treatment for argv
        g['argv']= ayrton.runner.globals['argv']

        logger.debug3 ('globals passed to remote: %s', ayrton.utils.dump_dict (g))
        global_env= pickle.dumps (g)
        logger.debug3 ('locals passed to remote: %s', ayrton.utils.dump_dict (l))
        local_env= pickle.dumps (l)

        port= 4227

        if not self._debug and not self._test:
            precommand= ''
        else:
            precommand= '''import os; os.chdir (%r)''' % os.getcwd ()

        command= '''python3 -c "#!                                        #  1
import pickle                                                             #  2
# names needed for unpickling                                             #  3
from ast import Module, Assign, Name, Store, Call, Load, Expr             #  4
import sys                                                                #  5
from socket import socket                                                 #  6
import ayrton #  this means that ayrton has to be installed in the remote #  7
import traceback                                                          #  8
                                                                          #  9
import logging                                                            # 10
logger= logging.getLogger ('ayrton.remote.runner')                        # 11
                                                                          # 12
# precommand, used by tests to change to the proper directory             # 13
# so it picks up the current version of the code.                         # 14
%s                                                                        # 15
                                                                          # 16
client= socket ()                                                         # 17
client.connect (('127.0.0.1', %d))                                        # 18
ast= pickle.loads (client.recv (%d))                                      # 19
logger.debug ('code to run:\\n%%s', ayrton.ast_pprinter.pprint (ast))     # 20
g= pickle.loads (client.recv (%d))                                        # 21
logger.debug2 ('globals received: %%s', ayrton.utils.dump_dict (g))       # 22
l= pickle.loads (client.recv (%d))                                        # 23
logger.debug2 ('locals received: %%s', ayrton.utils.dump_dict (l))        # 24
                                                                          # 25
# set the global runner so functions and Commands work                    # 26
ayrton.runner= ayrton.Ayrton (g, l)                                       # 27
caught= None                                                              # 28
result= None                                                              # 29
                                                                          # 30
try:                                                                      # 31
    result= ayrton.runner.run_tree (ast, 'from_remote')                   # 32
except Exception as e:                                                    # 33
    logger.debug ('run raised: %%r', e)                                   # 34
    logger.debug (traceback.format_exc())                                 # 35
    caught= e                                                             # 36
                                                                          # 37
logger.debug2 ('runner.locals: %%s', ayrton.utils.dump_dict (ayrton.runner.locals)) # 38
                                                                          # 39
logger.debug ('about to send exit status')                                # 40
data = pickle.dumps ( (ayrton.runner.locals, result, caught) )            # 41
logger.debug ('sending %%d bytes', len (data))                            # 42
client.sendall (data)                                                     # 43
logger.debug ('exit status sent')                                         # 44
client.close ()                                                           # 45"
''' % (precommand, port, len (self.ast), len (global_env), len (local_env))

        logger.debug ('code to execute remote: %s', command)

        if not self._debug:
            self.client= paramiko.SSHClient ()
            # TODO: TypeError: invalid file: ['/home/mdione/.ssh/known_hosts']
            # self.client.load_host_keys (bash ('~/.ssh/known_hosts'))
            # self.client.set_missing_host_key_policy (ShutUpPolicy ())
            self.client.set_missing_host_key_policy (paramiko.WarningPolicy ())
            self.client.connect (self.hostname, *self.args, **self.kwargs)

            # create the backchannel
            # this channel will be used for sending/receiving runtime data
            # to/from the remote
            # the remote code will connect to it (line #18)
            # read the ast (#19), globals (#21) and locals (#23)
            # and return the locals, result and exception (#43)
            # the remote will see this channel as a localhost port
            # and it's seen on the local side as self.con defined below
            self.result_listen= self.client.get_transport ()
            self.result_listen.request_port_forward ('localhost', port)

            # taken from paramiko/client.py:SSHClient.exec_command()
            channel= self.client.get_transport ().open_session ()
            # TODO:
            #19:44:54.953791 getsockopt(3, SOL_TCP, TCP_NODELAY, [0], [4]) = 0 <0.000016>
            #19:44:54.953852 setsockopt(3, SOL_TCP, TCP_NODELAY, [1], 4) = 0 <0.000014>

            try:
                # TODO signal handler from SIGWINCH
                term= shutil.get_terminal_size ()
                channel.get_pty (os.environ['TERM'], term.columns, term.lines)
            except OSError:
                channel.get_pty (os.environ['TERM'], )

            channel.exec_command (command)
            i= o= e= channel

            self.result_channel= self.result_listen.accept ()
        else:
            self.client= socket ()
            self.client.connect ((self.hostname, 2233)) # nc listening here, see DebugRemoteTests
            # unbuffered
            i= open (self.client.fileno (), 'wb', 0)
            o= open (self.client.fileno (), 'rb', 0)
            e= open (self.client.fileno (), 'rb', 0)

            self.result_listen= socket ()
            # self.result_listen.setsockopt (SO_REUSEADDR, )
            self.result_listen.bind (('', port))
            self.result_listen.listen (1)

            # so bash does not hang waiting from more input
            command+= 'exit\n'
            i.write (command.encode ())

            (self.result_channel, addr)= self.result_listen.accept ()

        logger.debug ('sending ast, globals, locals')
        self.result_channel.sendall (self.ast)
        self.result_channel.sendall (global_env)
        self.result_channel.sendall (local_env)

        # TODO: handle _in, _out, _err
        self.remote= RemoteStub (( (os.dup (0), i), (o, os.dup (1)), (e, os.dup (2)) ))


    def __exit__ (self, *args):
        logger.debug (args)
        data= b''
        partial= self.result_channel.recv (8196)
        while len(partial)>0:
            data+= partial
            partial= self.result_channel.recv (8196)

        logger.debug ('recieved %d bytes', len (data))
        (l, result, e)= pickle.loads (data)
        logger.debug ('result from remote: %r', result)
        logger.debug3 ('locals returned from remote: %s', ayrton.utils.dump_dict (l))
        logger.debug ('closing %s', self.result_channel)
        self.result_channel.close ()

        logger.debug ('closing %s', self.result_listen)
        self.result_listen.close ()
        logger.debug ('closing %s', self.remote)
        self.remote.close ()

        # update locals
        callers_frame= sys._getframe().f_back
        logger.debug3 ('caller name: %s', callers_frame.f_code.co_name)
        callers_frame.f_locals.update (l)
        # see https://mail.python.org/pipermail/python-dev/2005-January/051018.html
        ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(callers_frame), 0)
        if self._debug:
            # this makes sure that remote locals were properly store in the fast locals
            ctypes.pythonapi.PyFrame_FastToLocals(ctypes.py_object(callers_frame))

        # TODO: (and globals?)

        logger.debug3 ('globals after remote: %s', ayrton.utils.dump_dict (ayrton.runner.globals))
        logger.debug3 ('locals after remote: %s', ayrton.utils.dump_dict (callers_frame.f_locals))
        logger.debug3 ('co_varnames: %s', callers_frame.f_code.co_varnames)

        if e is not None:
            logger.debug ('raised from remote: %r', e)
            # TODO: this makes the exception be as if raised from here
            raise e



from ayrton.parser.astcompiler import consts, misc
from ayrton.parser import error
from ayrton.parser.pyparser.pygram import syms, tokens
from ayrton.parser.pyparser.error import SyntaxError
from ayrton.execute import Command
import ast

def ast_from_node(space, node, compile_info):
    """Turn a parse tree, node, to AST."""
    return ASTBuilder(space, node, compile_info).build_ast()


augassign_operator_map = {
    '+='  : ast.Add,
    '-='  : ast.Sub,
    '/='  : ast.Div,
    '//=' : ast.FloorDiv,
    '%='  : ast.Mod,
    '<<='  : ast.LShift,
    '>>='  : ast.RShift,
    '&='  : ast.BitAnd,
    '|='  : ast.BitOr,
    '^='  : ast.BitXor,
    '*='  : ast.Mult,
    '**=' : ast.Pow
}

operator_map = misc.dict_to_switch({
    tokens.VBAR : ast.BitOr,
    tokens.CIRCUMFLEX : ast.BitXor,
    tokens.AMPER : ast.BitAnd,
    tokens.LEFTSHIFT : ast.LShift,
    tokens.RIGHTSHIFT : ast.RShift,
    tokens.PLUS : ast.Add,
    tokens.MINUS : ast.Sub,
    tokens.STAR : ast.Mult,
    tokens.SLASH : ast.Div,
    tokens.DOUBLESLASH : ast.FloorDiv,
    tokens.PERCENT : ast.Mod
})

expression_name_map = {
    ast.Attribute: 'attr',
    ast.Name: 'id',
}


def parsestr(space, encoding, literal):
    # return space.wrap(eval (literal))
    return eval (literal)

class ASTBuilder(object):

    def __init__(self, space, n, compile_info):
        self.space = space
        self.compile_info = compile_info
        self.root_node = n

    def build_ast(self):
        """Convert an top level parse tree node into an AST mod."""
        n = self.root_node
        if n.type == syms.file_input:
            stmts = []
            for i in range(len(n.children) - 1):
                stmt = n.children[i]
                if stmt.type == tokens.NEWLINE:
                    continue
                sub_stmts_count = self.number_of_statements(stmt)
                if sub_stmts_count == 1:
                    stmts.append(self.handle_stmt(stmt))
                else:
                    stmt = stmt.children[0]
                    for j in range(sub_stmts_count):
                        small_stmt = stmt.children[j * 2]
                        stmts.append(self.handle_stmt(small_stmt))
            return ast.Module(stmts)
        elif n.type == syms.eval_input:
            body = self.handle_testlist(n.children[0])
            return ast.Expression(body)
        elif n.type == syms.single_input:
            first_child = n.children[0]
            if first_child.type == tokens.NEWLINE:
                # An empty line.
                return ast.Interactive([])
            else:
                num_stmts = self.number_of_statements(first_child)
                if num_stmts == 1:
                    stmts = [self.handle_stmt(first_child)]
                else:
                    stmts = []
                    for i in range(0, len(first_child.children), 2):
                        stmt = first_child.children[i]
                        if stmt.type == tokens.NEWLINE:
                            break
                        stmts.append(self.handle_stmt(stmt))
                return ast.Interactive(stmts)
        else:
            raise AssertionError("unknown root node")

    def number_of_statements(self, n):
        """Compute the number of AST statements contained in a node."""
        stmt_type = n.type
        if stmt_type == syms.compound_stmt:
            return 1
        elif stmt_type == syms.stmt:
            return self.number_of_statements(n.children[0])
        elif stmt_type == syms.simple_stmt:
            # Divide to remove semi-colons.
            return len(n.children) // 2
        else:
            raise AssertionError("non-statement node")

    def error(self, msg, n):
        """Raise a SyntaxError with the lineno and column set to n's."""
        raise SyntaxError(msg, n.lineno, n.col_offset,
                          filename=self.compile_info.filename)

    def error_ast(self, msg, ast_node):
        raise SyntaxError(msg, ast_node.lineno, ast_node.col_offset,
                          filename=self.compile_info.filename)

    def check_forbidden_name(self, name, node):
        try:
            misc.check_forbidden_name(name)
        except misc.ForbiddenNameAssignment as e:
            self.error("cannot assign to %s" % (e.name,), node)

    def new_identifier(self, name):
        return misc.new_identifier(self.space, name)

    def set_context(self, expr, ctx):
        """Set the context of an expression to Store or Del if possible."""
        t = type(expr)
        try:
            # TODO: check if Starred is ok
            if t in (ast.Attribute, ast.Name):
                if type(ctx) == ast.Store():
                    mis.check_forbidden_name(getattr (expr, expression_name_map[t]), expr)
            elif t in (ast.Subscript, ast.Starred):
                pass
            elif t in (ast.List, ast.Tuple):
                for elt in expr.elts:
                    self.set_context(elt, ctx)
            expr.ctx = ctx
        except misc.ForbiddenNameAssignment as e:
            self.error_ast("cannot assign to %s" % (e.name,), e.node)

    def handle_del_stmt(self, del_node):
        targets = self.handle_exprlist(del_node.children[1], ast.Del())
        new_node = ast.Delete (targets)
        new_node.lineno = ( del_node.lineno)
        new_node.col_offset = del_node.col_offset
        return new_node

    def handle_flow_stmt(self, flow_node):
        first_child = flow_node.children[0]
        first_child_type = first_child.type
        if first_child_type == syms.break_stmt:
            new_node = ast.Break ()
            new_node.lineno = flow_node.lineno
            new_node.col_offset = flow_node.col_offset
            return new_node
        elif first_child_type == syms.continue_stmt:
            new_node = ast.Continue ()
            new_node.lineno = flow_node.lineno
            new_node.col_offset = flow_node.col_offset
            return new_node
        elif first_child_type == syms.yield_stmt:
            yield_expr = self.handle_expr(first_child.children[0])
            new_node = ast.Expr (yield_expr)
            new_node.lineno = flow_node.lineno
            new_node.col_offset = flow_node.col_offset
            return new_node
        elif first_child_type == syms.return_stmt:
            if len(first_child.children) == 1:
                values = None
            else:
                values = self.handle_testlist(first_child.children[1])
            new_node = ast.Return (values)
            new_node.lineno = flow_node.lineno
            new_node.col_offset = flow_node.col_offset
            return new_node
        elif first_child_type == syms.raise_stmt:
            exc = None
            cause = None
            child_count = len(first_child.children)
            if child_count >= 2:
                exc = self.handle_expr(first_child.children[1])
            if child_count >= 4:
                cause = self.handle_expr(first_child.children[3])
            new_node = ast.Raise (exc, cause)
            new_node.lineno = flow_node.lineno
            new_node.col_offset = flow_node.col_offset
            return new_node
        else:
            raise AssertionError("unknown flow statement")

    def alias_for_import_name(self, import_name, store=True):
        while True:
            import_name_type = import_name.type
            if import_name_type == syms.import_as_name:
                name = self.new_identifier(import_name.children[0].value)
                if len(import_name.children) == 3:
                    as_name = self.new_identifier(
                        import_name.children[2].value)
                    self.check_forbidden_name(as_name, import_name.children[2])
                else:
                    as_name = None
                    self.check_forbidden_name(name, import_name.children[0])
                return ast.alias(name, as_name)
            elif import_name_type == syms.dotted_as_name:
                if len(import_name.children) == 1:
                    import_name = import_name.children[0]
                    continue
                alias = self.alias_for_import_name(import_name.children[0],
                                                   store=False)
                asname_node = import_name.children[2]
                alias.asname = self.new_identifier(asname_node.value)
                self.check_forbidden_name(alias.asname, asname_node)
                return alias
            elif import_name_type == syms.dotted_name:
                if len(import_name.children) == 1:
                    name = self.new_identifier(import_name.children[0].value)
                    if store:
                        self.check_forbidden_name(name, import_name.children[0])
                    return ast.alias(name, None)
                name_parts = [import_name.children[i].value
                              for i in range(0, len(import_name.children), 2)]
                name = ".".join(name_parts)
                return ast.alias(name, None)
            elif import_name_type == tokens.STAR:
                return ast.alias("*", None)
            else:
                raise AssertionError("unknown import name")

    def handle_import_stmt(self, import_node):
        import_node = import_node.children[0]
        if import_node.type == syms.import_name:
            dotted_as_names = import_node.children[1]
            aliases = [self.alias_for_import_name(dotted_as_names.children[i])
                       for i in range(0, len(dotted_as_names.children), 2)]
            new_node = ast.Import (aliases)
            new_node.lineno = import_node.lineno
            new_node.col_offset = import_node.col_offset
            return new_node
        elif import_node.type == syms.import_from:
            child_count = len(import_node.children)
            module = None
            modname = None
            i = 1
            dot_count = 0
            while i < child_count:
                child = import_node.children[i]
                child_type = child.type
                if child_type == syms.dotted_name:
                    module = self.alias_for_import_name(child, False)
                    i += 1
                    break
                elif child_type == tokens.ELLIPSIS:
                    # Special case for tokenization.
                    dot_count += 2
                elif child_type != tokens.DOT:
                    break
                i += 1
                dot_count += 1
            i += 1
            after_import_type = import_node.children[i].type
            star_import = False
            if after_import_type == tokens.STAR:
                names_node = import_node.children[i]
                star_import = True
            elif after_import_type == tokens.LPAR:
                names_node = import_node.children[i + 1]
            elif after_import_type == syms.import_as_names:
                names_node = import_node.children[i]
                if len(names_node.children) % 2 == 0:
                    self.error("trailing comma is only allowed with "
                               "surronding parenthesis", names_node)
            else:
                raise AssertionError("unknown import node")
            if star_import:
                aliases = [self.alias_for_import_name(names_node)]
            else:
                aliases = [self.alias_for_import_name(names_node.children[i])
                           for i in range(0, len(names_node.children), 2)]
            if module is not None:
                modname = module.name
            new_node = ast.ImportFrom (modname, aliases, dot_count)
            new_node.lineno = import_node.lineno
            new_node.col_offset = import_node.col_offset
            return new_node
        else:
            raise AssertionError("unknown import node")

    def handle_global_stmt(self, global_node):
        names = [self.new_identifier(global_node.children[i].value)
                 for i in range(1, len(global_node.children), 2)]
        new_node = ast.Global (names)
        new_node.lineno = global_node.lineno
        new_node.col_offset = global_node.col_offset
        return new_node

    def handle_nonlocal_stmt(self, nonlocal_node):
        names = [self.new_identifier(nonlocal_node.children[i].value)
                 for i in range(1, len(nonlocal_node.children), 2)]
        new_node = ast.Nonlocal (names)
        new_node.lineno = nonlocal_node.lineno
        new_node.col_offset = nonlocal_node.col_offset
        return new_node

    def handle_assert_stmt(self, assert_node):
        expr = self.handle_expr(assert_node.children[1])
        msg = None
        if len(assert_node.children) == 4:
            msg = self.handle_expr(assert_node.children[3])
        new_node = ast.Assert (expr, msg)
        new_node.lineno = assert_node.lineno
        new_node.col_offset = assert_node.col_offset
        return new_node

    def handle_suite(self, suite_node):
        first_child = suite_node.children[0]
        if first_child.type == syms.simple_stmt:
            end = len(first_child.children) - 1
            if first_child.children[end - 1].type == tokens.SEMI:
                end -= 1
            stmts = [self.handle_stmt(first_child.children[i])
                     for i in range(0, end, 2)]
        else:
            stmts = []
            for i in range(2, len(suite_node.children) - 1):
                stmt = suite_node.children[i]
                stmt_count = self.number_of_statements(stmt)
                if stmt_count == 1:
                    stmts.append(self.handle_stmt(stmt))
                else:
                    simple_stmt = stmt.children[0]
                    for j in range(0, len(simple_stmt.children), 2):
                        stmt = simple_stmt.children[j]
                        if not stmt.children:
                            break
                        stmts.append(self.handle_stmt(stmt))
        return stmts

    def handle_if_stmt(self, if_node):
        child_count = len(if_node.children)
        if child_count == 4:
            test = self.handle_expr(if_node.children[1])
            suite = self.handle_suite(if_node.children[3])
            new_node = ast.If (test, suite, [])
            new_node.lineno = if_node.lineno
            new_node.col_offset = if_node.col_offset
            return new_node
        otherwise_string = if_node.children[4].value
        if otherwise_string == "else":
            test = self.handle_expr(if_node.children[1])
            suite = self.handle_suite(if_node.children[3])
            else_suite = self.handle_suite(if_node.children[6])
            new_node = ast.If (test, suite, else_suite)
            new_node.lineno = if_node.lineno
            new_node.col_offset = if_node.col_offset
            return new_node
        elif otherwise_string == "elif":
            elif_count = child_count - 4
            after_elif = if_node.children[elif_count + 1]
            if after_elif.type == tokens.NAME and \
                    after_elif.value == "else":
                has_else = True
                elif_count -= 3
            else:
                has_else = False
            elif_count /= 4
            if has_else:
                last_elif = if_node.children[-6]
                last_elif_test = self.handle_expr(last_elif)
                elif_body = self.handle_suite(if_node.children[-4])
                else_body = self.handle_suite(if_node.children[-1])
                new_node = ast.If(last_elif_test, elif_body, else_body)
                new_node.lineno = last_elif.lineno
                new_node.col_offset = last_elif.col_offset
                otherwise = [new_node]
                elif_count -= 1
            else:
                otherwise = []
            for i in range(elif_count):
                offset = 5 + (elif_count - i - 1) * 4
                elif_test_node = if_node.children[offset]
                elif_test = self.handle_expr(elif_test_node)
                elif_body = self.handle_suite(if_node.children[offset + 2])
                new_if = ast.If (elif_test, elif_body, otherwise)
                new_if.lineno = elif_test_node.lineno
                new_if.col_offset = elif_test_node.col_offset
                otherwise = [new_if]
            expr = self.handle_expr(if_node.children[1])
            body = self.handle_suite(if_node.children[3])
            new_node = ast.If (expr, body, otherwise)
            new_node.lineno = if_node.lineno
            new_node.col_offset = if_node.col_offset
            return new_node
        else:
            raise AssertionError("unknown if statement configuration")

    def handle_while_stmt(self, while_node):
        loop_test = self.handle_expr(while_node.children[1])
        body = self.handle_suite(while_node.children[3])
        if len(while_node.children) == 7:
            otherwise = self.handle_suite(while_node.children[6])
        else:
            otherwise = []
        new_node = ast.While (loop_test, body, otherwise)
        new_node.lineno = while_node.lineno
        new_node.col_offset = while_node.col_offset
        return new_node

    def handle_for_stmt(self, for_node):
        target_node = for_node.children[1]
        target_as_exprlist = self.handle_exprlist(target_node, ast.Store())
        if len(target_node.children) == 1:
            target = target_as_exprlist[0]
        else:
            target = ast.Tuple (target_as_exprlist, ast.Store())
            target.lineno = target_node.lineno
            target.col_offset = target_node.col_offset
        expr = self.handle_testlist(for_node.children[3])
        body = self.handle_suite(for_node.children[5])
        if len(for_node.children) == 9:
            otherwise = self.handle_suite(for_node.children[8])
        else:
            otherwise = []
        new_node = ast.For (target, expr, body, otherwise)
        new_node.lineno = for_node.lineno
        new_node.col_offset = for_node.col_offset
        return new_node

    def handle_except_clause(self, exc, body):
        test = None
        name = None
        suite = self.handle_suite(body)
        child_count = len(exc.children)
        if child_count >= 2:
            test = self.handle_expr(exc.children[1])
        if child_count == 4:
            name_node = exc.children[3]
            name = self.new_identifier(name_node.value)
            self.check_forbidden_name(name, name_node)
        new_node = ast.ExceptHandler (test, name, suite)
        new_node.lineno = exc.lineno
        new_node.col_offset = exc.col_offset
        return new_node

    def handle_try_stmt(self, try_node):
        body = self.handle_suite(try_node.children[2])
        child_count = len(try_node.children)
        except_count = (child_count - 3 ) // 3
        otherwise = []
        finally_suite = []
        possible_extra_clause = try_node.children[-3]
        if possible_extra_clause.type == tokens.NAME:
            if possible_extra_clause.value == "finally":
                if child_count >= 9 and \
                        try_node.children[-6].type == tokens.NAME:
                    otherwise = self.handle_suite(try_node.children[-4])
                    except_count -= 1
                finally_suite = self.handle_suite(try_node.children[-1])
                except_count -= 1
            else:
                otherwise = self.handle_suite(try_node.children[-1])
                except_count -= 1
        handlers = []
        if except_count:
            for i in range(except_count):
                base_offset = i * 3
                exc = try_node.children[3 + base_offset]
                except_body = try_node.children[5 + base_offset]
                handlers.append(self.handle_except_clause(exc, except_body))
        new_node = ast.Try (body, handlers, otherwise, finally_suite)
        new_node.lineno = try_node.lineno
        new_node.col_offset = try_node.col_offset
        return new_node

    def handle_with_stmt(self, with_node):
        body = self.handle_suite(with_node.children[-1])
        i = len(with_node.children) - 1
        while True:
            i -= 2
            item = with_node.children[i]
            test = self.handle_expr(item.children[0])
            if len(item.children) == 3:
                target = self.handle_expr(item.children[2])
                self.set_context(target, ast.Store())
            else:
                target = None
            wi = ast.With (test, target, body)
            wi.lineno = with_node.lineno
            wi.col_offset = with_node.col_offset
            if i == 1:
                break
            body = [wi]
        return wi

    def handle_with_item(self, item_node):
        test = self.handle_expr(item_node.children[0])
        if len(item_node.children) == 3:
            target = self.handle_expr(item_node.children[2])
            self.set_context(target, ast.Store())
        else:
            target = None
        return ast.withitem(test, target)

    def handle_with_stmt(self, with_node):
        body = self.handle_suite(with_node.children[-1])
        items = [self.handle_with_item(with_node.children[i])
                 for i in range(1, len(with_node.children)-2, 2)]
        new_node = ast.With (items, body)
        new_node.lineno = with_node.lineno
        new_node.col_offset = with_node.col_offset
        return new_node

    def handle_classdef(self, classdef_node, decorators=None):
        if decorators is None:
            decorators = []
        name_node = classdef_node.children[1]
        name = self.new_identifier(name_node.value)
        self.check_forbidden_name(name, name_node)
        if len(classdef_node.children) == 4:
            # class NAME ':' suite
            body = self.handle_suite(classdef_node.children[3])
            new_node = ast.ClassDef (name, [], [], None, None, body, decorators)
            new_node.lineno = classdef_node.lineno
            new_node.col_offset = classdef_node.col_offset
            return new_node
        if classdef_node.children[3].type == tokens.RPAR:
            # class NAME '(' ')' ':' suite
            body = self.handle_suite(classdef_node.children[5])
            new_node = ast.ClassDef (name, [], [], None, None, body, decorators)
            new_node.lineno = classdef_node.lineno
            new_node.col_offset = classdef_node.col_offset
            return new_node

        # class NAME '(' arglist ')' ':' suite
        # build up a fake Call node so we can extract its pieces
        # what it's doing is use handle_call() to parse the 'arguments'
        # see ClassDef(name='Foo', bases=[], keywords=[], starargs=None, kwargs=None, body=[Pass()], decorator_list=[])
        # and handle_suite() to parse the body
        call_name = ast.Name (name, ast.Load())
        call_name.lineno = classdef_node.lineno
        call_name.col_offset = classdef_node.col_offset
        call = self.handle_call(classdef_node.children[3], call_name)
        body = self.handle_suite(classdef_node.children[6])
        new_node = ast.ClassDef (name, call.args, call.keywords, call.starargs, call.kwargs, body, decorators)
        new_node.lineno = classdef_node.lineno
        new_node.col_offset = classdef_node.col_offset
        return new_node

    def handle_class_bases(self, bases_node):
        if len(bases_node.children) == 1:
            return [self.handle_expr(bases_node.children[0])]
        return self.get_expression_list(bases_node)

    def handle_funcdef(self, funcdef_node, decorators=None):
        if decorators is None:
            decorators = []
        name_node = funcdef_node.children[1]
        name = self.new_identifier(name_node.value)
        self.check_forbidden_name(name, name_node)
        args = self.handle_arguments(funcdef_node.children[2])
        suite = 4
        returns = None
        if funcdef_node.children[3].type == tokens.RARROW:
            returns = self.handle_expr(funcdef_node.children[4])
            suite += 2
        body = self.handle_suite(funcdef_node.children[suite])
        new_node = ast.FunctionDef (name, args, body, decorators, returns)
        new_node.lineno = funcdef_node.lineno
        new_node.col_offset = funcdef_node.col_offset
        return new_node

    def handle_decorated(self, decorated_node):
        decorators = self.handle_decorators(decorated_node.children[0])
        definition = decorated_node.children[1]
        if definition.type == syms.funcdef:
            node = self.handle_funcdef(definition, decorators)
        elif definition.type == syms.classdef:
            node = self.handle_classdef(definition, decorators)
        else:
            raise AssertionError("unkown decorated")
        node.lineno = decorated_node.lineno
        node.col_offset = decorated_node.col_offset
        return node

    def handle_decorators(self, decorators_node):
        return [self.handle_decorator(dec) for dec in decorators_node.children]

    def handle_decorator(self, decorator_node):
        dec_name = self.handle_dotted_name(decorator_node.children[1])
        if len(decorator_node.children) == 3:
            dec = dec_name
        elif len(decorator_node.children) == 5:
            dec = ast.Call (dec_name, None, None, None, None)
            dec.lineno = decorator_node.lineno
            dec.col_offset = decorator_node.col_offset
        else:
            dec = self.handle_call(decorator_node.children[3], dec_name)
        return dec

    def handle_dotted_name(self, dotted_name_node):
        base_value = self.new_identifier(dotted_name_node.children[0].value)
        name = ast.Name (base_value, ast.Load())
        name.lineno = dotted_name_node.lineno
        name.col_offset = dotted_name_node.col_offset
        for i in range(2, len(dotted_name_node.children), 2):
            attr = dotted_name_node.children[i].value
            attr = self.new_identifier(attr)
            name = ast.Attribute (name, attr, ast.Load())
            name.lineno = dotted_name_node.lineno
            name.col_offset = dotted_name_node.col_offset
        return name

    def handle_arguments(self, arguments_node):
        # This function handles both typedargslist (function definition)
        # and varargslist (lambda definition).
        if arguments_node.type == syms.parameters:
            if len(arguments_node.children) == 2:
                return ast.arguments([], None, [], [], None, [])
            arguments_node = arguments_node.children[1]
        i = 0
        child_count = len(arguments_node.children)
        n_pos = 0
        n_pos_def = 0
        n_kwdonly = 0
        # scan args
        while i < child_count:
            arg_type = arguments_node.children[i].type
            if arg_type == tokens.STAR:
                i += 1
                if i < child_count:
                    next_arg_type = arguments_node.children[i].type
                    if (next_arg_type == syms.tfpdef or
                        next_arg_type == syms.vfpdef):
                        i += 1
                break
            if arg_type == tokens.DOUBLESTAR:
                break
            if arg_type == syms.vfpdef or arg_type == syms.tfpdef:
                n_pos += 1
            if arg_type == tokens.EQUAL:
                n_pos_def += 1
            i += 1
        while i < child_count:
            arg_type = arguments_node.children[i].type
            if arg_type == tokens.DOUBLESTAR:
                break
            if arg_type == syms.vfpdef or arg_type == syms.tfpdef:
                n_kwdonly += 1
            i += 1
        pos = []
        posdefaults = []
        kwonly = []
        kwdefaults = []
        kwarg = None
        kwargann = None
        vararg = None
        varargann = None
        if n_pos + n_kwdonly > 255:
            self.error("more than 255 arguments", arguments_node)
        # process args
        i = 0
        have_default = False
        while i < child_count:
            arg = arguments_node.children[i]
            arg_type = arg.type
            if arg_type == syms.tfpdef or arg_type == syms.vfpdef:
                if i + 1 < child_count and \
                        arguments_node.children[i + 1].type == tokens.EQUAL:
                    default_node = arguments_node.children[i + 2]
                    posdefaults.append(self.handle_expr(default_node))
                    i += 2
                    have_default = True
                elif have_default:
                    msg = "non-default argument follows default argument"
                    self.error(msg, arguments_node)
                pos.append(self.handle_arg(arg))
                i += 2
            elif arg_type == tokens.STAR:
                if i + 1 >= child_count:
                    self.error("named arguments must follow bare *",
                               arguments_node)
                name_node = arguments_node.children[i + 1]
                keywordonly_args = []
                if name_node.type == tokens.COMMA:
                    i += 2
                    i = self.handle_keywordonly_args(arguments_node, i, kwonly,
                                                     kwdefaults)
                else:
                    vararg = name_node.children[0].value
                    vararg = self.new_identifier(vararg)
                    self.check_forbidden_name(vararg, name_node)
                    if len(name_node.children) > 1:
                        varargann = self.handle_expr(name_node.children[2])
                    i += 3
                    if i < child_count:
                        next_arg_type = arguments_node.children[i].type
                        if (next_arg_type == syms.tfpdef or
                            next_arg_type == syms.vfpdef):
                            i = self.handle_keywordonly_args(arguments_node, i,
                                                             kwonly, kwdefaults)
            elif arg_type == tokens.DOUBLESTAR:
                name_node = arguments_node.children[i + 1]
                kwarg = name_node.children[0].value
                kwarg = self.new_identifier(kwarg)
                self.check_forbidden_name(kwarg, name_node)
                if len(name_node.children) > 1:
                    kwargann = self.handle_expr(name_node.children[2])
                i += 3
            else:
                raise AssertionError("unknown node in argument list")
        # TODO: fix annotations
        # def __init__(self, args, vararg, varargannotation, kwonlyargs, kwarg, kwargannotation, defaults, kw_defaults):
        # arguments(args=[arg(arg='a', annotation=None), arg(arg='b', annotation=None)],
        #           vararg=arg(arg='more', annotation=None), kwonlyargs=[], kw_defaults=[],
        #           kwarg=arg(arg='kmore', annotation=None), defaults=[Num(n=1)])
        return ast.arguments(pos, vararg, kwonly, kwdefaults, kwarg, posdefaults)

    def handle_keywordonly_args(self, arguments_node, i, kwonly, kwdefaults):
        if kwonly is None:
            self.error("named arguments must follows bare *",
                       arguments_node.children[i])
        child_count = len(arguments_node.children)
        while i < child_count:
            arg = arguments_node.children[i]
            arg_type = arg.type
            if arg_type == syms.vfpdef or arg_type == syms.tfpdef:
                if (i + 1 < child_count and
                    arguments_node.children[i + 1].type == tokens.EQUAL):
                    expr = self.handle_expr(arguments_node.children[i + 2])
                    kwdefaults.append(expr)
                    i += 2
                else:
                    kwdefaults.append(None)
                ann = None
                if len(arg.children) == 3:
                    ann = self.handle_expr(arg.children[2])
                name_node = arg.children[0]
                argname = name_node.value
                argname = self.new_identifier(argname)
                self.check_forbidden_name(argname, name_node)
                kwonly.append(ast.arg(argname, ann))
                i += 2
            elif arg_type == tokens.DOUBLESTAR:
                return i
        return i

    def handle_arg(self, arg_node):
        name_node = arg_node.children[0]
        name = self.new_identifier(name_node.value)
        self.check_forbidden_name(name, arg_node)
        ann = None
        if len(arg_node.children) == 3:
            ann = self.handle_expr(arg_node.children[2])
        return ast.arg(name, ann)

    def handle_stmt(self, stmt):
        stmt_type = stmt.type
        if stmt_type == syms.stmt:
            stmt = stmt.children[0]
            stmt_type = stmt.type
        if stmt_type == syms.simple_stmt:
            stmt = stmt.children[0]
            stmt_type = stmt.type
        if stmt_type == syms.small_stmt:
            stmt = stmt.children[0]
            stmt_type = stmt.type
            if stmt_type == syms.expr_stmt:
                return self.handle_expr_stmt(stmt)
            elif stmt_type == syms.del_stmt:
                return self.handle_del_stmt(stmt)
            elif stmt_type == syms.pass_stmt:
                new_node = ast.Pass ()
                new_node.lineno = stmt.lineno
                new_node.col_offset = stmt.col_offset
                return new_node
            elif stmt_type == syms.flow_stmt:
                return self.handle_flow_stmt(stmt)
            elif stmt_type == syms.import_stmt:
                return self.handle_import_stmt(stmt)
            elif stmt_type == syms.global_stmt:
                return self.handle_global_stmt(stmt)
            elif stmt_type == syms.nonlocal_stmt:
                return self.handle_nonlocal_stmt(stmt)
            elif stmt_type == syms.assert_stmt:
                return self.handle_assert_stmt(stmt)
            else:
                raise AssertionError("unhandled small statement")
        elif stmt_type == syms.compound_stmt:
            stmt = stmt.children[0]
            stmt_type = stmt.type
            if stmt_type == syms.if_stmt:
                return self.handle_if_stmt(stmt)
            elif stmt_type == syms.while_stmt:
                return self.handle_while_stmt(stmt)
            elif stmt_type == syms.for_stmt:
                return self.handle_for_stmt(stmt)
            elif stmt_type == syms.try_stmt:
                return self.handle_try_stmt(stmt)
            elif stmt_type == syms.with_stmt:
                return self.handle_with_stmt(stmt)
            elif stmt_type == syms.funcdef:
                return self.handle_funcdef(stmt)
            elif stmt_type == syms.classdef:
                return self.handle_classdef(stmt)
            elif stmt_type == syms.decorated:
                return self.handle_decorated(stmt)
            else:
                raise AssertionError("unhandled compound statement")
        else:
            raise AssertionError("unknown statment type")

    def handle_expr_stmt(self, stmt):
        if len(stmt.children) == 1:
            expression = self.handle_testlist(stmt.children[0])
            new_node = ast.Expr (expression)
            new_node.lineno = stmt.lineno
            new_node.col_offset = stmt.col_offset
            return new_node
        elif stmt.children[1].type == syms.augassign:
            # Augmented assignment.
            target_child = stmt.children[0]
            target_expr = self.handle_testlist(target_child)
            self.set_context(target_expr, ast.Store())
            value_child = stmt.children[2]
            if value_child.type == syms.testlist:
                value_expr = self.handle_testlist(value_child)
            else:
                value_expr = self.handle_expr(value_child)
            op_str = stmt.children[1].children[0].value
            operator = augassign_operator_map[op_str]
            new_node = ast.AugAssign (target_expr, operator(), value_expr)
            new_node.lineno = stmt.lineno
            new_node.col_offset = stmt.col_offset
            return new_node
        else:
            # Normal assignment.
            targets = []
            for i in range(0, len(stmt.children) - 2, 2):
                target_node = stmt.children[i]
                if target_node.type == syms.yield_expr:
                    self.error("assignment to yield expression not possible",
                               target_node)
                target_expr = self.handle_testlist(target_node)
                self.set_context(target_expr, ast.Store())
                targets.append(target_expr)
            value_child = stmt.children[-1]
            if value_child.type == syms.testlist_star_expr:
                value_expr = self.handle_testlist(value_child)
            else:
                value_expr = self.handle_expr(value_child)
            new_node = ast.Assign (targets, value_expr)
            new_node.lineno = stmt.lineno
            new_node.col_offset = stmt.col_offset
            return new_node

    def get_expression_list(self, tests):
        return [self.handle_expr(tests.children[i])
                for i in range(0, len(tests.children), 2)]

    def handle_testlist(self, tests):
        if len(tests.children) == 1:
            return self.handle_expr(tests.children[0])
        else:
            elts = self.get_expression_list(tests)
            new_node = ast.Tuple (elts, ast.Load())
            new_node.lineno = tests.lineno
            new_node.col_offset = tests.col_offset
            return new_node

    def handle_expr(self, expr_node):
        # Loop until we return something.
        while True:
            expr_node_type = expr_node.type
            if expr_node_type == syms.test or expr_node_type == syms.test_nocond:
                first_child = expr_node.children[0]
                if first_child.type in (syms.lambdef, syms.lambdef_nocond):
                    return self.handle_lambdef(first_child)
                elif len(expr_node.children) > 1:
                    return self.handle_ifexp(expr_node)
                else:
                    expr_node = first_child
            elif expr_node_type == syms.or_test or \
                    expr_node_type == syms.and_test:
                if len(expr_node.children) == 1:
                    expr_node = expr_node.children[0]
                    continue
                seq = [self.handle_expr(expr_node.children[i])
                       for i in range(0, len(expr_node.children), 2)]
                if expr_node_type == syms.or_test:
                    op = ast.Or
                else:
                    op = ast.And
                new_node = ast.BoolOp (op(), seq)
                new_node.lineno = expr_node.lineno
                new_node.col_offset = expr_node.col_offset
                return new_node
            elif expr_node_type == syms.not_test:
                if len(expr_node.children) == 1:
                    expr_node = expr_node.children[0]
                    continue
                expr = self.handle_expr(expr_node.children[1])
                new_node = ast.UnaryOp (ast.Not(), expr)
                new_node.lineno = expr_node.lineno
                new_node.col_offset = expr_node.col_offset
                return new_node
            elif expr_node_type == syms.comparison:
                if len(expr_node.children) == 1:
                    expr_node = expr_node.children[0]
                    continue
                operators = []
                operands = []
                expr = self.handle_expr(expr_node.children[0])
                for i in range(1, len(expr_node.children), 2):
                    operators.append(self.handle_comp_op(expr_node.children[i]))
                    operands.append(self.handle_expr(expr_node.children[i + 1]))
                new_node = ast.Compare (expr, operators, operands)
                new_node.lineno = expr_node.lineno
                new_node.col_offset = expr_node.col_offset
                return new_node
            elif expr_node_type == syms.star_expr:
                return self.handle_star_expr(expr_node)
            elif expr_node_type == syms.expr or \
                    expr_node_type == syms.xor_expr or \
                    expr_node_type == syms.and_expr or \
                    expr_node_type == syms.shift_expr or \
                    expr_node_type == syms.arith_expr or \
                    expr_node_type == syms.term:
                if len(expr_node.children) == 1:
                    expr_node = expr_node.children[0]
                    continue
                return self.handle_binop(expr_node)
            elif expr_node_type == syms.yield_expr:
                is_from = False
                if len(expr_node.children) > 1:
                    arg_node = expr_node.children[1]  # yield arg
                    if len(arg_node.children) == 2:
                        is_from = True
                        expr = self.handle_expr(arg_node.children[1])
                    else:
                        expr = self.handle_testlist(arg_node.children[0])
                else:
                    expr = None
                if is_from:
                    new_node = ast.YieldFrom (expr)
                    new_node.lineno = expr_node.lineno
                    new_node.col_offset = expr_node.col_offset
                    return new_node
                new_node = ast.Yield (expr)
                new_node.lineno = expr_node.lineno
                new_node.col_offset = expr_node.col_offset
                return new_node
            elif expr_node_type == syms.factor:
                if len(expr_node.children) == 1:
                    expr_node = expr_node.children[0]
                    continue
                return self.handle_factor(expr_node)
            elif expr_node_type == syms.power:
                return self.handle_power(expr_node)
            else:
                raise AssertionError("unknown expr")

    def handle_star_expr(self, star_expr_node):
        expr = self.handle_expr(star_expr_node.children[1])
        new_node = ast.Starred (expr, ast.Load())
        new_node.lineno = star_expr_node.lineno
        new_node.col_offset = star_expr_node.col_offset
        return new_node

    def handle_lambdef(self, lambdef_node):
        expr = self.handle_expr(lambdef_node.children[-1])
        if len(lambdef_node.children) == 3:
            args = ast.arguments([], None, [], [], None, [])
        else:
            args = self.handle_arguments(lambdef_node.children[1])
        new_node = ast.Lambda (args, expr)
        new_node.lineno = lambdef_node.lineno
        new_node.col_offset = lambdef_node.col_offset
        return new_node

    def handle_ifexp(self, if_expr_node):
        body = self.handle_expr(if_expr_node.children[0])
        expression = self.handle_expr(if_expr_node.children[2])
        otherwise = self.handle_expr(if_expr_node.children[4])
        new_node = ast.IfExp (expression, body, otherwise)
        new_node.lineno = if_expr_node.lineno
        new_node.col_offset = if_expr_node.col_offset
        return new_node

    def handle_comp_op(self, comp_op_node):
        comp_node = comp_op_node.children[0]
        comp_type = comp_node.type
        if len(comp_op_node.children) == 1:
            if comp_type == tokens.LESS:
                return ast.Lt()
            elif comp_type == tokens.GREATER:
                return ast.Gt()
            elif comp_type == tokens.EQEQUAL:
                return ast.Eq()
            elif comp_type == tokens.LESSEQUAL:
                return ast.LtE()
            elif comp_type == tokens.GREATEREQUAL:
                return ast.GtE()
            elif comp_type == tokens.NOTEQUAL:
                flufl = self.compile_info.flags & consts.CO_FUTURE_BARRY_AS_BDFL
                if flufl and comp_node.value == '!=':
                    self.error('invalid comparison', comp_node)
                elif not flufl and comp_node.value == '<>':
                    self.error('invalid comparison', comp_node)
                return ast.NotEq()
            elif comp_type == tokens.NAME:
                if comp_node.value == "is":
                    return ast.Is()
                elif comp_node.value == "in":
                    return ast.In()
                else:
                    raise AssertionError("invalid comparison")
            else:
                raise AssertionError("invalid comparison")
        else:
            if comp_op_node.children[1].value == "in":
                return ast.NotIn()
            elif comp_node.value == "is":
                return ast.IsNot()
            else:
                raise AssertionError("invalid comparison")

    def handle_binop(self, binop_node):
        left = self.handle_expr(binop_node.children[0])
        right = self.handle_expr(binop_node.children[2])
        op = operator_map(binop_node.children[1].type)
        result = ast.BinOp (left, op(), right)
        result.lineno = binop_node.lineno
        result.col_offset = binop_node.col_offset
        number_of_ops = (len(binop_node.children) - 1) // 2
        for i in range(1, number_of_ops):
            op_node = binop_node.children[i * 2 + 1]
            op = operator_map(op_node.type)
            sub_right = self.handle_expr(binop_node.children[i * 2 + 2])
            result = ast.BinOp (result, op(), sub_right)
            result.lineno = op_node.lineno
            result.col_offset = op_node.col_offset
        return result

    def handle_factor(self, factor_node):
        expr = self.handle_expr(factor_node.children[1])
        op_type = factor_node.children[0].type
        if op_type == tokens.PLUS:
            op = ast.UAdd
        elif op_type == tokens.MINUS:
            op = ast.USub
        elif op_type == tokens.TILDE:
            op = ast.Invert
        else:
            raise AssertionError("invalid factor node")
        new_node = ast.UnaryOp (op(), expr)
        new_node.lineno = factor_node.lineno
        new_node.col_offset = factor_node.col_offset
        return new_node

    def handle_power(self, power_node):
        atom_expr = self.handle_atom(power_node.children[0])
        if len(power_node.children) == 1:
            return atom_expr
        for i in range(1, len(power_node.children)):
            trailer = power_node.children[i]
            if trailer.type != syms.trailer:
                break
            tmp_atom_expr = self.handle_trailer(trailer, atom_expr)
            tmp_atom_expr.lineno = atom_expr.lineno
            tmp_atom_expr.col_offset = atom_expr.col_offset
            atom_expr = tmp_atom_expr
        if power_node.children[-1].type == syms.factor:
            right = self.handle_expr(power_node.children[-1])
            atom_expr = ast.BinOp (atom_expr, ast.Pow(), right)
            atom_expr.lineno = power_node.lineno
            atom_expr.col_offset = power_node.col_offset
        return atom_expr

    def handle_slice(self, slice_node):
        first_child = slice_node.children[0]
        if len(slice_node.children) == 1 and first_child.type == syms.test:
            index = self.handle_expr(first_child)
            return ast.Index(index)
        lower = None
        upper = None
        step = None
        if first_child.type == syms.test:
            lower = self.handle_expr(first_child)
        if first_child.type == tokens.COLON:
            if len(slice_node.children) > 1:
                second_child = slice_node.children[1]
                if second_child.type == syms.test:
                    upper = self.handle_expr(second_child)
        elif len(slice_node.children) > 2:
            third_child = slice_node.children[2]
            if third_child.type == syms.test:
                upper = self.handle_expr(third_child)
        last_child = slice_node.children[-1]
        if last_child.type == syms.sliceop:
            if len(last_child.children) != 1:
                step_child = last_child.children[1]
                if step_child.type == syms.test:
                    step = self.handle_expr(step_child)
        return ast.Slice(lower, upper, step)

    def handle_trailer(self, trailer_node, left_expr):
        first_child = trailer_node.children[0]
        if first_child.type == tokens.LPAR:
            if len(trailer_node.children) == 2:
                new_node = ast.Call (left_expr, [], [], None, None)
                new_node.lineno = trailer_node.lineno
                new_node.col_offset = trailer_node.col_offset
                return new_node
            else:
                return self.handle_call(trailer_node.children[1], left_expr)
        elif first_child.type == tokens.DOT:
            attr = self.new_identifier(trailer_node.children[1].value)
            new_node = ast.Attribute (left_expr, attr, ast.Load())
            new_node.lineno = trailer_node.lineno
            new_node.col_offset = trailer_node.col_offset
            return new_node
        else:
            middle = trailer_node.children[1]
            if len(middle.children) == 1:
                slice = self.handle_slice(middle.children[0])
                new_node = ast.Subscript (left_expr, slice, ast.Load())
                new_node.lineno = middle.lineno
                new_node.col_offset = middle.col_offset
                return new_node
            slices = []
            simple = True
            for i in range(0, len(middle.children), 2):
                slc = self.handle_slice(middle.children[i])
                if not isinstance(slc, ast.Index):
                    simple = False
                slices.append(slc)
            if not simple:
                ext_slice = ast.ExtSlice(slices)
                new_node = ast.Subscript (left_expr, ext_slice, ast.Load())
                new_node.lineno = middle.lineno
                new_node.col_offset = middle.col_offset
                return new_node
            elts = []
            for idx in slices:
                assert isinstance(idx, ast.Index)
                elts.append(idx.value)
            tup = ast.Tuple (elts, ast.Load())
            tup.lineno = middle.lineno
            tup.col_offset = middle.col_offset
            new_node = ast.Subscript(left_expr, ast.Index (tup), ast.Load())
            new_node.lineno = middle.lineno
            new_node.col_offset = middle.col_offset
            return new_node

    def handle_call(self, args_node, callable_expr):
        arg_count = 0
        keyword_count = 0
        generator_count = 0
        for argument in args_node.children:
            if argument.type == syms.argument:
                if len(argument.children) == 1:
                    arg_count += 1
                elif argument.children[1].type == syms.comp_for:
                    generator_count += 1
                else:
                    keyword_count += 1
        if generator_count > 1 or \
                (generator_count and (keyword_count or arg_count)):
            self.error("Generator expression must be parenthesized "
                       "if not sole argument", args_node)
        if arg_count + keyword_count + generator_count > 255:
            self.error("more than 255 arguments", args_node)
        args = []
        keywords = []
        variable_arg = None
        keywords_arg = None
        child_count = len(args_node.children)
        i = 0
        while i < child_count:
            argument = args_node.children[i]
            if argument.type == syms.argument:
                if len(argument.children) == 1:
                    expr_node = argument.children[0]
                    if variable_arg:
                        self.error("only named arguments may follow "
                                   "*expression", expr_node)
                    args.append(self.handle_expr(expr_node))
                elif argument.children[1].type == syms.comp_for:
                    args.append(self.handle_genexp(argument))
                else:
                    keyword_node = argument.children[0]
                    keyword_expr = self.handle_expr(keyword_node)
                    if isinstance(keyword_expr, ast.Lambda):
                        self.error("lambda cannot contain assignment",
                                   keyword_node)
                    keyword = keyword_expr
                    if isinstance (keyword, ast.Name):
                        self.check_forbidden_name(keyword.id, keyword_node)
                    keyword_value = self.handle_expr(argument.children[2])
                    if isinstance (keyword, ast.Name) and keyword.id in Command.supported_options:
                        keywords.append(ast.keyword(keyword.id, keyword_value))
                    else:
                        kw = ast.keyword(keyword, keyword_value)
                        kw.lineno = keyword_node.lineno
                        kw.col_offset = keyword_node.col_offset
                        name = ast.Name ('o', ast.Load())
                        name.lineno = keyword_node.lineno
                        name.col_offset = keyword_node.col_offset
                        arg = ast.Call(name, [], [ kw ], None, None)
                        arg.lineno = keyword_node.lineno
                        arg.col_offset = keyword.col_offset
                        args.append(arg)
            elif argument.type == tokens.STAR:
                variable_arg = self.handle_expr(args_node.children[i + 1])
                i += 1
            elif argument.type == tokens.DOUBLESTAR:
                keywords_arg = self.handle_expr(args_node.children[i + 1])
                i += 1
            i += 1
        if not args:
            args = []
        if not keywords:
            keywords = []
        new_node = ast.Call(callable_expr, args, keywords, variable_arg, keywords_arg)
        new_node.lineno = callable_expr.lineno
        new_node.col_offset = callable_expr.col_offset
        return new_node

    def parse_number(self, raw):
        return eval(raw)

    def handle_atom(self, atom_node):
        first_child = atom_node.children[0]
        first_child_type = first_child.type
        if first_child_type == tokens.NAME:
            name = self.new_identifier(first_child.value)
            new_node = ast.Name (name, ast.Load())
            new_node.lineno = first_child.lineno
            new_node.col_offset = first_child.col_offset
            return new_node
        elif first_child_type == tokens.STRING:
            space = self.space
            encoding = self.compile_info.encoding
            try:
                sub_strings_w = [parsestr(space, encoding, s.value)
                                 for s in atom_node.children]
            except error.OperationError as e:
                if not (e.match(space, space.w_UnicodeError) or
                        e.match(space, space.w_ValueError)):
                    raise
                # Unicode/ValueError in literal: turn into SyntaxError
                self.error(e.errorstr(space), atom_node)
                sub_strings_w = [] # please annotator
            # Implement implicit string concatenation.
            w_string = sub_strings_w[0]
            for i in range(1, len(sub_strings_w)):
                try:
                    w_string = space.add(w_string, sub_strings_w[i])
                except error.OperationError as e:
                    if not e.match(space, space.w_TypeError):
                        raise
                    self.error("cannot mix bytes and nonbytes literals",
                              atom_node)
                # UnicodeError in literal: turn into SyntaxError
            strdata = type(w_string)==str
            node_cls = ast.Str if strdata else ast.Bytes
            new_node = node_cls(w_string)
            new_node.lineno = atom_node.lineno
            new_node.col_offset = atom_node.col_offset
            return new_node
        elif first_child_type == tokens.NUMBER:
            num_value = self.parse_number(first_child.value)
            new_node = ast.Num (num_value)
            new_node.lineno = atom_node.lineno
            new_node.col_offset = atom_node.col_offset
            return new_node
        elif first_child_type == tokens.ELLIPSIS:
            new_node = ast.Ellipsis ()
            new_node.lineno = atom_node.lineno
            new_node.col_offset = atom_node.col_offset
            return new_node
        elif first_child_type == tokens.LPAR:
            second_child = atom_node.children[1]
            if second_child.type == tokens.RPAR:
                new_node = ast.Tuple (None, ast.Load())
                new_node.lineno = atom_node.lineno
                new_node.col_offset = atom_node.col_offset
                return new_node
            elif second_child.type == syms.yield_expr:
                return self.handle_expr(second_child)
            return self.handle_testlist_gexp(second_child)
        elif first_child_type == tokens.LSQB:
            second_child = atom_node.children[1]
            if second_child.type == tokens.RSQB:
                new_node = ast.List ([], ast.Load())
                new_node.lineno = atom_node.lineno
                new_node.col_offset = atom_node.col_offset
                return new_node
            if len(second_child.children) == 1 or \
                    second_child.children[1].type == tokens.COMMA:
                elts = self.get_expression_list(second_child)
                new_node = ast.List (elts, ast.Load())
                new_node.lineno = atom_node.lineno
                new_node.col_offset = atom_node.col_offset
                return new_node
            return self.handle_listcomp(second_child)
        elif first_child_type == tokens.LBRACE:
            maker = atom_node.children[1]
            if maker.type == tokens.RBRACE:
                new_node = ast.Dict (None, None)
                new_node.lineno = atom_node.lineno
                new_node.col_offset = atom_node.col_offset
                return new_node
            n_maker_children = len(maker.children)
            if n_maker_children == 1 or maker.children[1].type == tokens.COMMA:
                elts = []
                for i in range(0, n_maker_children, 2):
                    elts.append(self.handle_expr(maker.children[i]))
                new_node = ast.Set (elts)
                new_node.lineno = atom_node.lineno
                new_node.col_offset = atom_node.col_offset
                return new_node
            if maker.children[1].type == syms.comp_for:
                return self.handle_setcomp(maker)
            if (n_maker_children > 3 and
                maker.children[3].type == syms.comp_for):
                return self.handle_dictcomp(maker)
            keys = []
            values = []
            for i in range(0, n_maker_children, 4):
                keys.append(self.handle_expr(maker.children[i]))
                values.append(self.handle_expr(maker.children[i + 2]))
            new_node = ast.Dict (keys, values)
            new_node.lineno = atom_node.lineno
            new_node.col_offset = atom_node.col_offset
            return new_node
        else:
            raise AssertionError("unknown atom")

    def handle_testlist_gexp(self, gexp_node):
        if len(gexp_node.children) > 1 and \
                gexp_node.children[1].type == syms.comp_for:
            return self.handle_genexp(gexp_node)
        return self.handle_testlist(gexp_node)

    def count_comp_fors(self, comp_node):
        count = 0
        current_for = comp_node
        while True:
            count += 1
            if len(current_for.children) == 5:
                current_iter = current_for.children[4]
            else:
                return count
            while True:
                first_child = current_iter.children[0]
                if first_child.type == syms.comp_for:
                    current_for = current_iter.children[0]
                    break
                elif first_child.type == syms.comp_if:
                    if len(first_child.children) == 3:
                        current_iter = first_child.children[2]
                    else:
                        return count
                else:
                    raise AssertionError("should not reach here")

    def count_comp_ifs(self, iter_node):
        count = 0
        while True:
            first_child = iter_node.children[0]
            if first_child.type == syms.comp_for:
                return count
            count += 1
            if len(first_child.children) == 2:
                return count
            iter_node = first_child.children[2]

    def comprehension_helper(self, comp_node):
        fors_count = self.count_comp_fors(comp_node)
        comps = []
        for i in range(fors_count):
            for_node = comp_node.children[1]
            for_targets = self.handle_exprlist(for_node, ast.Store())
            expr = self.handle_expr(comp_node.children[3])
            assert isinstance(expr, ast.expr)
            if len(for_node.children) == 1:
                comp = ast.comprehension(for_targets[0], expr, [])
            else:
                # Modified in python2.7, see http://bugs.python.org/issue6704
                # Fixing unamed tuple location
                expr_node = for_targets[0]
                assert isinstance(expr_node, ast.expr)
                col = expr_node.col_offset
                line = expr_node.lineno
                target = ast.Tuple(for_targets, ast.Store(), line, col)
                comp = ast.comprehension(target, expr, [])
            if len(comp_node.children) == 5:
                comp_node = comp_iter = comp_node.children[4]
                assert comp_iter.type == syms.comp_iter
                ifs_count = self.count_comp_ifs(comp_iter)
                if ifs_count:
                    ifs = []
                    for j in range(ifs_count):
                        comp_node = comp_if = comp_iter.children[0]
                        ifs.append(self.handle_expr(comp_if.children[1]))
                        if len(comp_if.children) == 3:
                            comp_node = comp_iter = comp_if.children[2]
                    comp.ifs = ifs
                if comp_node.type == syms.comp_iter:
                    comp_node = comp_node.children[0]
            assert isinstance(comp, ast.comprehension)
            comps.append(comp)
        return comps

    def handle_genexp(self, genexp_node):
        elt = self.handle_expr(genexp_node.children[0])
        comps = self.comprehension_helper(genexp_node.children[1])
        new_node = ast.GeneratorExp (elt, comps)
        new_node.lineno = genexp_node.lineno
        new_node.col_offset = genexp_node.col_offset
        return new_node

    def handle_listcomp(self, listcomp_node):
        elt = self.handle_expr(listcomp_node.children[0])
        comps = self.comprehension_helper(listcomp_node.children[1])
        new_node = ast.ListComp (elt, comps)
        new_node.lineno = listcomp_node.lineno
        new_node.col_offset = listcomp_node.col_offset
        return new_node

    def handle_setcomp(self, set_maker):
        elt = self.handle_expr(set_maker.children[0])
        comps = self.comprehension_helper(set_maker.children[1])
        new_node = ast.SetComp (elt, comps)
        new_node.lineno = set_maker.lineno
        new_node.col_offset = set_maker.col_offset
        return new_node

    def handle_dictcomp(self, dict_maker):
        key = self.handle_expr(dict_maker.children[0])
        value = self.handle_expr(dict_maker.children[2])
        comps = self.comprehension_helper(dict_maker.children[3])
        new_node = ast.DictComp (key, value, comps)
        new_node.lineno = dict_maker.lineno
        new_node.col_offset = dict_maker.col_offset
        return new_node

    def handle_exprlist(self, exprlist, context):
        exprs = []
        for i in range(0, len(exprlist.children), 2):
            child = exprlist.children[i]
            expr = self.handle_expr(child)
            self.set_context(expr, context)
            exprs.append(expr)
        return exprs




class SyntaxError(Exception):
    """Base class for exceptions raised by the parser."""

    def __init__(self, msg, lineno=0, offset=0, text=None, filename=None,
                 lastlineno=0):
        self.msg = msg
        self.lineno = lineno
        self.offset = offset
        self.text = text
        self.filename = filename
        self.lastlineno = lastlineno

    def wrap_info(self, space):
        w_text = w_filename = space.w_None
        if self.text is not None:
            from rpython.rlib.runicode import str_decode_utf_8
            # self.text may not be UTF-8 in case of decoding errors
            w_text = space.wrap(str_decode_utf_8(self.text, len(self.text),
                                                 'replace')[0])
        if self.filename is not None:
            w_filename = space.fsdecode(space.wrapbytes(self.filename))
        return space.newtuple([space.wrap(self.msg),
                               space.newtuple([w_filename,
                                               space.wrap(self.lineno),
                                               space.wrap(self.offset),
                                               w_text,
                                               space.wrap(self.lastlineno)])])

    def __str__(self):
        return "%s at pos (%d, %d) in %r" % (self.__class__.__name__,
                                             self.lineno,
                                             self.offset,
                                             self.text)

class IndentationError(SyntaxError):
    pass

class TabError(IndentationError):
    def __init__(self, lineno=0, offset=0, text=None, filename=None,
                 lastlineno=0):
        msg = "inconsistent use of tabs and spaces in indentation"
        IndentationError.__init__(self, msg, lineno, offset, text, filename, lastlineno)

class ASTError(Exception):
    def __init__(self, msg, ast_node ):
        self.msg = msg
        self.ast_node = ast_node


class TokenError(SyntaxError):

    def __init__(self, msg, line, lineno, column, tokens, lastlineno=0):
        SyntaxError.__init__(self, msg, lineno, column, line,
                             lastlineno=lastlineno)
        self.tokens = tokens

class TokenIndentationError(IndentationError):

    def __init__(self, msg, line, lineno, column, tokens):
        SyntaxError.__init__(self, msg, lineno, column, line)
        self.tokens = tokens



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# (c) 2013 Marcos Dione <mdione@grulic.org.ar>
# for licensing details see the file LICENSE.txt

from distutils.core import setup
import ayrton

setup (
    name='ayrton',
    version= ayrton.__version__,
    description= 'a shell-like scripting language based on Python3.',
    author= 'Marcos Dione',
    author_email= 'mdione@grulic.org.ar',
    url= 'https://github.com/StyXman/ayrton',
    packages= [ 'ayrton', 'ayrton.parser', 'ayrton.parser.pyparser',
               'ayrton.parser.astcompiler' ],
    package_data= {
        'ayrton.parser.pyparser': [ 'data/Grammar3.3' ],
        },
    scripts= [ 'bin/ayrton' ],
    license= 'GPLv3',
    classifiers= [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Topic :: System',
        'Topic :: System :: Systems Administration',
        ],
    )



# -*- coding: utf-8 -*-

# (c) 2015 Marcos Dione <mdione@grulic.org.ar>

# This file is part of ayrton.
#
# ayrton is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ayrton is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ayrton.  If not, see <http://www.gnu.org/licenses/>.

import logging
import functools

# TODO: there's probably a better way to do this
def debug2 (self, msg, *args, **kwargs):
    if self.manager.disable>=logging.DEBUG2:
        return

    if logging.DEBUG2>=self.getEffectiveLevel ():
        self._log (logging.DEBUG2, msg, args, **kwargs)


def debug3 (self, msg, *args, **kwargs):
    if self.manager.disable>=logging.DEBUG3:
        return

    if logging.DEBUG3>=self.getEffectiveLevel ():
        self._log (logging.DEBUG3, msg, args, **kwargs)


def patch_logging ():
    # based on https://mail.python.org/pipermail/tutor/2007-August/056243.html
    logging.DEBUG2= 9
    logging.DEBUG3= 8

    logging.addLevelName (logging.DEBUG2, 'DEBUG2')
    logging.addLevelName (logging.DEBUG3, 'DEBUG3')

    logging.Logger.debug2= debug2
    logging.Logger.debug3= debug3


patch_logging ()


def any_comparator (a, b):
    try:
        if a==b:
            return 0
        elif a<b:
            return -1
        else:
            return 1
    except TypeError:
        return any_comparator (str (type (a)), str (type (b)))


def dump_dict (d, level=1):
    if d is not None:
        strings= []

        if level==0:
            strings.append ("{\n")
        for k in sorted (d.keys (), key=functools.cmp_to_key (any_comparator)):
            v= d[k]
            if type (v)!=dict:
                strings.append ("%s%r: %r,\n" % ( '    '*level, k, v ))
            else:
                strings.append ("%s%r: {\n"   % ( '    '*level, k))
                strings.extend (dump_dict (v, level+1))
                strings.append ("%s},\n"      % ( '    '*level, ))
        if level==0:
            strings.cappend ("}\n")

        return ''.join (strings)
    else:
        return 'None'



from ayrton.parser.error import OperationError
from ayrton.parser.pyparser import future, parser, pytokenizer, pygram, error
from ayrton.parser.astcompiler import consts


def recode_to_utf8(space, bytes, encoding):
    if encoding == 'utf-8':
        return bytes
    w_text = space.call_method(space.wrapbytes(bytes), "decode",
                               space.wrap(encoding))
    w_recoded = space.call_method(w_text, "encode", space.wrap("utf-8"))
    return space.bytes_w(w_recoded)

def _normalize_encoding(encoding):
    """returns normalized name for <encoding>

    see dist/src/Parser/tokenizer.c 'get_normal_name()'
    for implementation details / reference

    NOTE: for now, parser.suite() raises a MemoryError when
          a bad encoding is used. (SF bug #979739)
    """
    if encoding is None:
        return None
    # lower() + '_' / '-' conversion
    encoding = encoding.replace('_', '-').lower()
    if encoding == 'utf-8' or encoding.startswith('utf-8-'):
        return 'utf-8'
    for variant in ['latin-1', 'iso-latin-1', 'iso-8859-1']:
        if (encoding == variant or
            encoding.startswith(variant + '-')):
            return 'iso-8859-1'
    return encoding

def _check_for_encoding(s):
    eol = s.find('\n')
    if eol < 0:
        return _check_line_for_encoding(s)[0]
    enc, again = _check_line_for_encoding(s[:eol])
    if enc or not again:
        return enc
    eol2 = s.find('\n', eol + 1)
    if eol2 < 0:
        return _check_line_for_encoding(s[eol + 1:])[0]
    return _check_line_for_encoding(s[eol + 1:eol2])[0]


def _check_line_for_encoding(line):
    """returns the declared encoding or None"""
    i = 0
    for i in range(len(line)):
        if line[i] == '#':
            break
        if line[i] not in ' \t\014':
            return None, False  # Not a comment, don't read the second line.
    return pytokenizer.match_encoding_declaration(line[i:]), True


# shamelessly ripped off rpython to avoid dependency
def check_str0(fname):
    """A 'probe' to trigger a failure at translation time, if the
    string was not proved to not contain NUL characters."""
    assert '\x00' not in fname, "NUL byte in string"

class CompileInfo(object):
    """Stores information about the source being compiled.

    * filename: The filename of the source.
    * mode: The parse mode to use. ('exec', 'eval', or 'single')
    * flags: Parser and compiler flags.
    * encoding: The source encoding.
    * last_future_import: The line number and offset of the last __future__
      import.
    * hidden_applevel: Will this code unit and sub units be hidden at the
      applevel?
    * optimize: optimization level:
        -1 = same as interpreter,
         0 = no optmiziation,
         1 = remove asserts,
         2 = remove docstrings.
    """

    def __init__(self, filename, mode="exec", flags=0, future_pos=(0, 0),
                 hidden_applevel=False, optimize=-1):
        check_str0(filename)
        self.filename = filename
        self.mode = mode
        self.encoding = None
        self.flags = flags
        self.optimize = optimize
        self.last_future_import = future_pos
        self.hidden_applevel = hidden_applevel


_targets = {
'eval' : pygram.syms.eval_input,
'single' : pygram.syms.single_input,
'exec' : pygram.syms.file_input,
}

class PythonParser(parser.Parser):

    def __init__(self, space, future_flags=future.futureFlags_3_2,
                 grammar=pygram.python_grammar):
        parser.Parser.__init__(self, grammar)
        self.space = space
        self.future_flags = future_flags

    def parse_source(self, bytessrc, compile_info):
        """Main entry point for parsing Python source.

        Everything from decoding the source to tokenizing to building the parse
        tree is handled here.
        """
        # Detect source encoding.
        enc = None
        if compile_info.flags & consts.PyCF_SOURCE_IS_UTF8:
            enc = 'utf-8'

        if compile_info.flags & consts.PyCF_IGNORE_COOKIE:
            textsrc = bytessrc
        elif bytessrc.startswith("\xEF\xBB\xBF"):
            bytessrc = bytessrc[3:]
            enc = 'utf-8'
            # If an encoding is explicitly given check that it is utf-8.
            decl_enc = _check_for_encoding(bytessrc)
            if decl_enc and decl_enc != "utf-8":
                raise error.SyntaxError("UTF-8 BOM with %s coding cookie" % decl_enc,
                                        filename=compile_info.filename)
            textsrc = bytessrc
        else:
            enc = _normalize_encoding(_check_for_encoding(bytessrc))
            if enc is None:
                enc = 'utf-8'
            try:
                textsrc = recode_to_utf8(self.space, bytessrc, enc)
            except OperationError as e:
                # if the codec is not found, LookupError is raised.  we
                # check using 'is_w' not to mask potential IndexError or
                # KeyError
                space = self.space
                if e.match(space, space.w_LookupError):
                    raise error.SyntaxError("Unknown encoding: %s" % enc,
                                            filename=compile_info.filename)
                # Transform unicode errors into SyntaxError
                if e.match(space, space.w_UnicodeDecodeError):
                    e.normalize_exception(space)
                    w_message = space.str(e.get_w_value(space))
                    raise error.SyntaxError(space.str_w(w_message))
                raise

        flags = compile_info.flags

        # The tokenizer is very picky about how it wants its input.
        source_lines = textsrc.splitlines(True)
        if source_lines and not source_lines[-1].endswith("\n"):
            source_lines[-1] += '\n'
        if textsrc and textsrc[-1] == "\n":
            flags &= ~consts.PyCF_DONT_IMPLY_DEDENT

        self.prepare(_targets[compile_info.mode])
        tp = 0
        try:
            try:
                # Note: we no longer pass the CO_FUTURE_* to the tokenizer,
                # which is expected to work independently of them.  It's
                # certainly the case for all futures in Python <= 2.7.
                tokens = pytokenizer.generate_tokens(source_lines, flags)

                newflags, last_future_import = (
                    future.add_future_flags(self.future_flags, tokens))
                compile_info.last_future_import = last_future_import
                compile_info.flags |= newflags
                self.grammar = pygram.python_grammar
                tokens_stream = iter(tokens)

                for tp, value, lineno, column, line in tokens_stream:
                    if self.add_token(tp, value, lineno, column, line):
                        break

                if compile_info.mode == 'single':
                    for tp, value, lineno, column, line in tokens_stream:
                        if tp == pygram.tokens.ENDMARKER:
                            break
                        if tp == pygram.tokens.NEWLINE:
                            continue

                        if tp == pygram.tokens.COMMENT:
                            for tp, _, _, _, _ in tokens_stream:
                                if tp == pygram.tokens.NEWLINE:
                                    break
                        else:
                            new_err = error.SyntaxError
                            msg = ("multiple statements found while "
                                   "compiling a single statement")
                            raise new_err(msg, lineno, column,
                                          line, compile_info.filename)

            except error.TokenError as e:
                e.filename = compile_info.filename
                raise
            except parser.ParseError as e:
                # Catch parse errors, pretty them up and reraise them as a
                # SyntaxError.
                new_err = error.IndentationError
                if tp == pygram.tokens.INDENT:
                    msg = "unexpected indent"
                elif e.expected == pygram.tokens.INDENT:
                    msg = "expected an indented block"
                else:
                    new_err = error.SyntaxError
                    msg = "invalid syntax"
                raise new_err(msg, e.lineno, e.col_offset, e.line,
                              compile_info.filename)
            else:
                tree = self.root
        finally:
            # Avoid hanging onto the tree.
            self.root = None
        if enc is not None:
            compile_info.encoding = enc
        return tree
